#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT
"""
Constants for various key directory paths.
"""

import tempfile
from pathlib import Path

PACKAGE = Path(Path(__file__).resolve().parent)
"""The root directory of the package installation."""

DATA = Path(PACKAGE, r'data')
"""The root directory of all package data."""

CSS = Path(DATA, r'css')
"""The css directory."""

MCSS = Path(DATA, r'm.css')
"""The root directory of the bundled m.css build."""

GENERATED = Path(DATA, r'generated')
"""The root directory of all auto-generated package data."""

FONTS = Path(GENERATED, r'fonts')
"""Directory containing all the self-hosted google fonts."""

TESTS = Path(PACKAGE, r'..', r'tests')
"""The root directory of the repository's tests."""

TEMP = Path(tempfile.gettempdir(), r'poxy')
"""A global temp directory shared by all instances of poxy."""



#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

from .run import run
from .utils import lib_version, Error, WarningTreatedAsError
from .graph import GraphError, GraphNodePropertyChanged

__all__ = [r'run', r'lib_version', r'Error', r'WarningTreatedAsError', r'GraphError', r'GraphNodePropertyChanged']

__version__ = r'.'.join([str(v) for v in lib_version()])



#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT
"""
Functions for working with CSS files.
"""

from .utils import *
from . import dirs

RX_COMMENT = re.compile(r'''/[*].+?[*]/''', flags=re.DOTALL)
RX_IMPORT = re.compile(r'''@import\s+url\(\s*['"]?\s*(.+?)\s*['"]?\s*\)\s*;''', flags=re.I)
RX_MCSS_FILENAME = re.compile(r'(?:m|pygments)-[a-z0-9_-]+[.]css$', flags=re.I)
RX_GOOGLE_FONT = re.compile(
	r'''url\(\s*['"]?(https://fonts[.]gstatic[.]com/[a-z0-9_/%+?:-]+?[.]woff2)['"]?\s*\)''', flags=re.I
)



def strip_comments(text) -> str:
	global RX_COMMENT
	return RX_COMMENT.sub('', text)



def strip_quotes(text):
	if text.startswith(r'"') or text.startswith(r"'"):
		text = text[1:]
	if text.endswith(r'"') or text.endswith(r"'"):
		text = text[:-2]
	return text



def has_mcss_filename(path) -> bool:
	path = str(path).lower()
	if path.endswith(r'm-special.css'):
		return False

	global RX_MCSS_FILENAME
	return bool(RX_MCSS_FILENAME.match(path))



def resolve_imports(text, cwd=None, use_cached_fonts=True) -> typing.Tuple[str, bool]:
	if cwd is None:
		cwd = Path.cwd()
	cwd = coerce_path(cwd).resolve()
	assert_existing_directory(cwd)

	had_mcss_files = False

	def match_handler(m):
		global RX_MCSS_THEME
		nonlocal cwd
		nonlocal had_mcss_files
		nonlocal use_cached_fonts

		import_path = strip_quotes(m[1].strip())
		had_mcss_files = had_mcss_files or has_mcss_filename(import_path)
		path = None
		path_ok = lambda: path is not None and path.exists() and path.is_file()

		# download + cache uris locally
		if is_uri(import_path):
			dirs.GENERATED.mkdir(exist_ok=True)
			path = Path(dirs.GENERATED, rf'{sha1(import_path.lower())}.css')
			if not path_ok() or (not use_cached_fonts and str(import_path).find(r'font') != -1):
				print(rf"Downloading {import_path}")
				data = download_text(import_path)
				print(rf"Writing {path}")
				with open(path, 'w', encoding='utf-8', newline='\n') as f:
					f.write(data)

		# otherwise just check cwd
		if not path_ok():
			path = Path(cwd, import_path)

		# if we still haven't found a match just leave the @import statement as it was
		if not path_ok():
			return m[0]

		text = strip_comments(read_all_text_from_file(path, logger=True)).strip()
		header = rf'/*==== {import_path} {"="*(110-len(import_path))}*/'
		text = f'\n\n{header}\n{text}\n\n'

		res = resolve_imports(text, cwd=path.parent, use_cached_fonts=use_cached_fonts)
		had_mcss_files = had_mcss_files or res[1]
		return res[0]

	global RX_IMPORT
	return (RX_IMPORT.sub(match_handler, text), had_mcss_files)



def resolve_google_fonts(text, use_cached_fonts=True) -> str:

	def match_handler(m):
		global RX_GOOGLE_FONT
		uri = strip_quotes(m[1].strip())
		file_name = uri[uri.rfind('/') + 1:]
		dirs.FONTS.mkdir(exist_ok=True)
		path = Path(dirs.FONTS, rf'{file_name}')
		if not path.exists() or not use_cached_fonts:
			print(rf"Downloading {uri}")
			data = download_binary(uri)
			print(rf"Writing {path}")
			with open(path, 'wb') as f:
				f.write(data)

		return rf"url('fonts/{file_name}')"

	global RX_GOOGLE_FONT
	return RX_GOOGLE_FONT.sub(match_handler, text)



def minify(text) -> str:
	text = text.replace('\t', ' ')
	text = re.sub(r'\n[ \t]*[{]', ' {\n', text)
	text = re.sub(r'[ \t]+\n', '\n', text)
	text = re.sub(r'\n\n+', '\n', text)

	pos = 0
	open_brackets = []
	while pos < len(text):
		if text[pos:pos + 2] == r'/*':
			pos += 2
			while pos < len(text) - 1:
				if text[pos:pos + 2] == r'*/':
					pos += 2
					break
				pos += 1
		else:
			if text[pos] == r'{':
				open_brackets.append(pos)
			elif text[pos] == r'}':
				if open_brackets:
					block_content = text[open_brackets[-1] + 1:pos].strip().replace('\n', ' ')
					block_content = re.sub(r'[ \t][ \t]+', ' ', block_content)
					if r'{' not in block_content and r'}' not in block_content:
						semis = 0
						for c in block_content:
							if c == ';':
								semis += 1
							if semis >= 4:
								break
						if semis <= 4 and len(block_content) <= 100:
							new_block = rf'{{ {block_content} }}'
							text = text[:open_brackets[-1]] + new_block + text[pos + 1:]
							pos = open_brackets[-1] - 1 + len(new_block) - 1
					open_brackets.pop()
			pos += 1

	text = re.sub(r'[{]\s+?[}]', r'{}', text)
	text = re.sub(r'[ \t][ \t]+', ' ', text)
	text = re.sub(r'\n[ \t]+([^ \t])', r'\n\1', text)
	return text



def regenerate_builtin_styles(use_cached_fonts=True):
	dirs.GENERATED.mkdir(exist_ok=True)
	if not use_cached_fonts:
		delete_directory(dirs.FONTS, logger=True)
	THEMES = (Path(dirs.CSS, r'poxy.css'), )
	for theme_source_file in THEMES:
		text = strip_comments(read_all_text_from_file(theme_source_file, logger=True))
		text, had_mcss_files = resolve_imports(
			text,  #
			cwd=theme_source_file.parent,
			use_cached_fonts=use_cached_fonts,
		)
		text = resolve_google_fonts(text, use_cached_fonts=use_cached_fonts)
		text = re.sub(r':(before|after)', r'::\1', text)
		text = re.sub(r':::+(before|after)', r'::\1', text)
		text = text.replace('\r\n', '\n')
		text = text.replace('\r', '\n')
		text = minify(text)
		if had_mcss_files:
			mcss_license = read_all_text_from_file(Path(dirs.MCSS, 'COPYING'), logger=True).strip()
			text = rf'''/*
This file was automatically generated from multiple sources,
some of which included stylesheets from mosra/m.css.
The license for that project is as follows:

{mcss_license}
*/
{text}
'''
		theme_dest_file = Path(dirs.GENERATED, theme_source_file.name)
		print(rf'Writing {theme_dest_file}')
		with open(theme_dest_file, r'w', encoding=r'utf-8', newline='\n') as f:
			f.write(text)



#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT
"""
Helpers for working with HTML using BeautifulSoup.
"""

import bs4
from .utils import *

#=======================================================================================================================
# BS4 HELPER FUNCTIONS
#=======================================================================================================================



def find_parent(tag, names, cutoff=None):
	if not is_collection(names):
		names = (names, )
	parent = tag.parent
	while (parent is not None):
		if (cutoff is not None and parent is cutoff):
			return None
		if parent.name in names:
			return parent
		parent = parent.parent
	return parent



def destroy_node(node):
	assert node is not None
	if (isinstance(node, bs4.NavigableString)):
		node.extract()
	else:
		node.decompose()



def replace_tag(tag, new_tag_str):
	assert tag.parent is not None
	newTags = []
	if new_tag_str:
		doc = bs4.BeautifulSoup(new_tag_str, 'html5lib')
		if len(doc.body.contents) > 0:
			newTags = [f for f in doc.body.contents]
			newTags = [f.extract() for f in newTags]
			prev = tag
			for newTag in newTags:
				prev.insert_after(newTag)
				prev = newTag
	destroy_node(tag)
	return newTags



def shallow_search(starting_tag, names, filter=None):
	if isinstance(starting_tag, bs4.NavigableString):
		return []

	if not is_collection(names):
		names = (names, )

	if starting_tag.name in names:
		if filter is None or filter(starting_tag):
			return [starting_tag]

	results = []
	for tag in starting_tag.children:
		if isinstance(tag, bs4.NavigableString):
			continue
		if tag.name in names:
			if filter is None or filter(tag):
				results.append(tag)
		else:
			results = results + shallow_search(tag, names, filter)
	return results



def string_descendants(starting_tag, filter=None):
	if isinstance(starting_tag, bs4.NavigableString):
		if filter is None or filter(starting_tag):
			return [starting_tag]

	results = []
	for tag in starting_tag.children:
		if isinstance(tag, bs4.NavigableString):
			if filter is None or filter(tag):
				results.append(tag)
		else:
			results = results + string_descendants(tag, filter)
	return results



def add_class(tag, classes):
	assert tag is not None
	appended = False
	if 'class' not in tag.attrs:
		tag['class'] = []
	if not is_collection(classes):
		classes = (classes, )
	for class_ in classes:
		if class_ not in tag['class']:
			tag['class'].append(class_)
			appended = True
	return appended



def remove_class(tag, classes):
	assert tag is not None
	removed = False
	if 'class' in tag.attrs:
		if not is_collection(classes):
			classes = (classes, )
		for class_ in classes:
			if class_ in tag['class']:
				tag['class'].remove(class_)
				removed = True
		if removed and len(tag['class']) == 0:
			del tag['class']
	return removed



def set_class(tag, classes):
	assert tag is not None
	tag['class'] = []
	add_class(tag, classes)



def get_classes(tag) -> list:
	assert tag is not None
	if 'class' not in tag.attrs:
		return []
	classes = tag['class']
	if not is_collection(classes):
		classes = [classes]
	return [str(c) for c in classes]



def has_any_classes(tag, *classes) -> bool:
	assert tag is not None
	if 'class' not in tag.attrs:
		return False
	has = get_classes(tag)
	for cls in classes:
		if cls in has:
			return True
	return False



#=======================================================================================================================
# HTML DOCUMENT
#=======================================================================================================================



class HTMLDocument(object):

	def __init__(self, source: typing.Union[str, Path], logger):
		self.__logger = logger

		assert source is not None
		if isinstance(source, Path):
			with open(str(source), 'r', encoding='utf-8') as f:
				self.__doc = bs4.BeautifulSoup(f, 'html5lib', from_encoding='utf-8')
		else:
			self.__doc = bs4.BeautifulSoup(source, 'html5lib')

		self.html = self.__doc.html
		self.head = self.__doc.head
		self.body = self.__doc.body

		self.article = None
		self.article_content = None
		try:
			self.article = self.__doc.body.main.article
			self.article_content = self.article.div.div.div
		except:
			pass

		self.table_of_contents = None
		self.sections = None
		if self.article_content is not None:
			for toc_tag in ('nav', 'div'):
				for tag in self.article_content(toc_tag, class_='m-block m-default', recursive=False):
					if tag.h3 and tag.h3.string == 'Contents':
						self.table_of_contents = tag
						break
				if self.table_of_contents is not None:
					break
			self.sections = self.article_content('section', recursive=False)

	def smooth(self):
		self.__doc.smooth()

	def write(self, path):
		log(self.__logger, rf'Writing {path}')
		with open(self.path, 'w', encoding='utf-8', newline='\n') as f:
			f.write(str(self.__doc))

	def __str__(self) -> str:
		return str(self.__doc)

	def new_tag(self, tag_name, parent=None, string=None, class_=None, index=None, before=None, after=None, **kwargs):
		tag = self.__doc.new_tag(tag_name, **kwargs)
		if (string is not None):
			if (tag.string is not None):
				tag.string.replace_with(string)
			else:
				tag.string = bs4.NavigableString(string)
		if (class_ is not None):
			tag['class'] = class_
		if (before is not None):
			before.insert_before(tag)
		elif (after is not None):
			after.insert_after(tag)
		elif (parent is not None):
			if (index is None or index < 0):
				parent.append(tag)
			else:
				parent.insert(index, tag)

		return tag

	def find_all_from_sections(self, name=None, select=None, section=None, include_toc=False, **kwargs):
		tags = []
		if self.article_content is not None:
			sections = None
			if (section is not None):
				sections = self.article_content('section', recursive=False, id='section')
			else:
				sections = self.sections
			if include_toc and self.table_of_contents is not None:
				sections = [self.table_of_contents, *sections]
			for sect in sections:
				matches = sect(name, **kwargs) if name is not None else [sect]
				if (select is not None):
					newMatches = []
					for match in matches:
						newMatches += match.select(select)
					matches = newMatches
				tags += matches
		return tags



#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT
"""
The various entry-point methods used when poxy is invoked from the command line.
"""

import os
import sys
import argparse
import datetime
import subprocess
import zipfile
from subprocess import CalledProcessError
from schema import SchemaError
from .utils import *
from .run import run
from . import dirs
from . import css
from . import emoji
from . import mcss
from . import graph



def _invoker(func, **kwargs):
	try:
		func(**kwargs)
	except WarningTreatedAsError as err:
		print(rf'Error: {err} (warning treated as error)', file=sys.stderr)
		sys.exit(1)
	except graph.GraphError as err:
		print_exception(err, include_type=True, include_traceback=True, skip_frames=1)
		sys.exit(1)
	except SchemaError as err:
		print(err, file=sys.stderr)
		sys.exit(1)
	except Error as err:
		print(rf'Error: {err}', file=sys.stderr)
		sys.exit(1)
	except Exception as err:
		print('\n*************\n', file=sys.stderr)
		print_exception(err, include_type=True, include_traceback=True, skip_frames=1)
		print(
			'*************\n'
			'\nYou appear to have triggered an internal bug!'
			'\nPlease re-run poxy with --bug-report and file an issue at github.com/marzer/poxy/issues'
			'\nMany thanks!'
			'\n\n*************',
			file=sys.stderr
		)
		sys.exit(1)
	sys.exit(0)



def make_boolean_optional_arg(args, name, default, help='', **kwargs):
	if sys.version_info.minor >= 9:
		args.add_argument(rf'--{name}', default=default, help=help, action=argparse.BooleanOptionalAction, **kwargs)
	else:
		args.add_argument(rf'--{name}', action=r'store_true', help=help, **kwargs)
		args.add_argument(rf'--no-{name}', action=r'store_false', dest=name, **kwargs)
		args.set_defaults(**{name: default})



def main(invoker=True):
	"""
	The entry point when the library is invoked as `poxy`.
	"""
	if invoker:
		_invoker(main, invoker=False)
		return

	# yapf: disable
	args = argparse.ArgumentParser(
		description=
		r'''  _ __   _____  ___   _ ''' '\n'
		r''' | '_ \ / _ \ \/ / | | |''' '\n'
		r''' | |_) | (_) >  <| |_| |''' '\n'
		r''' | .__/ \___/_/\_\\__, |''' '\n'
		r''' | |               __/ |''' '\n'
		r''' |_|              |___/ ''' rf' v{lib_version_string()} - github.com/marzer/poxy'
		'\n\n'
		r'Generate fancy C++ documentation.',
		formatter_class=argparse.RawTextHelpFormatter
	)
	# yapf: enable

	#--------------------------------------------------------------
	# public user-facing arguments
	#--------------------------------------------------------------
	args.add_argument(
		r'config',
		type=Path,
		nargs='?',
		default=Path('.'),
		help=r'path to poxy.toml or a directory containing it (default: %(default)s)'
	)
	args.add_argument(
		r'-v',  #
		r'--verbose',
		action=r'store_true',
		help=r"enable very noisy diagnostic output"
	)
	make_boolean_optional_arg(
		args,  #
		r'html',
		default=True,
		help=r'specify whether HTML output is required'
	)
	args.add_argument(
		r'--ppinclude',  #
		type=str,
		default=None,
		metavar=r'<regex>',
		help=r"pattern matching HTML file names to post-process (default: all)"
	)
	args.add_argument(
		r'--ppexclude',  #
		type=str,
		default=None,
		metavar=r'<regex>',
		help=r"pattern matching HTML file names to exclude from post-processing (default: none)"
	)
	args.add_argument(
		r'--theme',  #
		choices=[r'light', r'dark', r'custom'],
		default=None,
		help=r'override the default visual theme (default: read from config)'
	)
	args.add_argument(
		r'--threads',  #
		type=int,
		default=0,
		metavar=r'N',
		help=r"set the number of threads to use (default: automatic)"
	)
	args.add_argument(
		r'--version',  #
		action=r'store_true',
		help=r"print the version and exit",
		dest=r'print_version'
	)
	make_boolean_optional_arg(
		args,  #
		r'xml',
		default=False,
		help=r'specify whether XML output is required'
	)
	make_boolean_optional_arg(
		args,  #
		r'werror',
		default=None,
		help=r'override the treating of warnings as errors (default: read from config)'
	)
	args.add_argument(
		r'--bug-report',  #
		action=r'store_true',
		help=r"captures all output in a zip file for easier bug reporting."
	)
	#--------------------------------------------------------------
	# hidden/developer-only/deprecated/diagnostic arguments
	#--------------------------------------------------------------
	args.add_argument(
		r'--nocleanup',  #
		action=r'store_true',
		help=argparse.SUPPRESS
	)
	args.add_argument(
		r'--noassets',  #
		action=r'store_true',
		help=argparse.SUPPRESS
	)
	args.add_argument(
		r'--update-styles',  #
		action=r'store_true',
		help=argparse.SUPPRESS
	)
	args.add_argument(
		r'--update-fonts',  #
		action=r'store_true',
		help=argparse.SUPPRESS
	)
	args.add_argument(
		r'--update-emoji',  #
		action=r'store_true',
		help=argparse.SUPPRESS
	)
	args.add_argument(
		r'--update-tests',  #
		action=r'store_true',
		help=argparse.SUPPRESS
	)
	args.add_argument(
		r'--update-mcss',  #
		type=Path,
		default=None,
		metavar=r'<path>',
		help=argparse.SUPPRESS,
		dest=r'mcss'
	)
	args.add_argument( # --xml and --html are the replacements for --xmlonly
		r'--xmlonly',  #
		action=r'store_true',
		help=argparse.SUPPRESS,
	)
	args.add_argument(
		r'--xml-v2',  #
		action=r'store_true',
		help=argparse.SUPPRESS
	)
	args.add_argument(
		r'--bug-report-internal',  #
		action=r'store_true',
		help=argparse.SUPPRESS
	)
	args = args.parse_args()

	#--------------------------------------------------------------
	# --version
	#--------------------------------------------------------------

	if args.print_version:
		print(r'.'.join([str(v) for v in lib_version()]))
		return

	#--------------------------------------------------------------
	# developer-only subcommands
	#--------------------------------------------------------------

	if args.mcss is not None:
		args.update_styles = True
		args.update_fonts = True
		mcss.update_bundled_install(args.mcss)
	assert_existing_directory(dirs.MCSS)
	assert_existing_file(Path(dirs.MCSS, r'documentation/doxygen.py'))

	if args.update_fonts:
		args.update_styles = True
	if args.update_styles:
		css.regenerate_builtin_styles(use_cached_fonts=not args.update_fonts)

	if args.update_emoji:
		emoji.update_database_file()

	if args.update_tests:
		if not dirs.TESTS.exists() or not dirs.TESTS.is_dir():
			raise Exception(
				f'{dirs.TESTS} did not exist or was not a directory (--update-tests is only for editable installations)'
			)
		run_python_script(
			Path(dirs.TESTS, r'regenerate_tests.py'),  #
			*[a for a in (r'--verbose' if args.verbose else None, r'--nocleanup' if args.nocleanup else None) if a]
		)

	if (args.update_styles or args.update_fonts or args.update_emoji or args.update_tests or args.mcss is not None):
		return

	#--------------------------------------------------------------
	# bug report invocation
	#--------------------------------------------------------------

	bug_report_directory = (Path.cwd() / r'poxy_bug_report').resolve()
	bug_report_zip = (Path.cwd() / r'poxy_bug_report.zip').resolve()

	if args.bug_report:
		bug_report_args = [arg for arg in sys.argv[1:] if arg not in (r'--bug-report', r'--bug-report-internal')]

		print(r'Preparing output paths')
		delete_directory(bug_report_directory)
		delete_file(bug_report_zip)
		os.makedirs(str(bug_report_directory), exist_ok=True)

		print(r'Invoking poxy')
		result = subprocess.run(
			args=[r'poxy', *bug_report_args, r'--bug-report-internal'],
			cwd=str(Path.cwd()),
			check=False,
			stdout=subprocess.PIPE,
			stderr=subprocess.STDOUT,
			encoding='utf-8'
		)

		if result.stdout is not None:
			print(r'Writing stdout')
			with open(bug_report_directory / r'stdout.txt', r'w', newline='\n', encoding=r'utf-8') as f:
				f.write(result.stdout)

		if result.stderr is not None:
			print(r'Writing stderr')
			with open(bug_report_directory / r'stderr.txt', r'w', newline='\n', encoding=r'utf-8') as f:
				f.write(result.stderr)

		print(r'Writing metadata')
		with open(bug_report_directory / r'metadata.txt', r'w', newline='\n', encoding=r'utf-8') as f:
			print(rf'version: {lib_version()}', file=f)
			print(rf'args: {bug_report_args}', file=f)
			print(rf'returncode: {result.returncode}', file=f)

		# zip file
		print(r'Zipping files')
		with zipfile.ZipFile(str(bug_report_zip), 'w', compression=zipfile.ZIP_DEFLATED, compresslevel=9) as zip:
			file_prefix_len = len(str(bug_report_directory))
			for file in get_all_files(bug_report_directory, recursive=True):
				if file.suffix is not None and file.suffix.lower() in (r'.pyc', ):
					continue
				relative_file = str(file)[file_prefix_len:].replace('\\', '/').strip('/')
				zip.write(file, arcname=rf'poxy_bug_report/{relative_file}')

		print(r'Cleaning up')
		delete_directory(bug_report_directory)

		print(
			f'Zip generated: {bug_report_zip}\n'
			'Please attach this file when you make a report at github.com/marzer/poxy/issues, thanks!'
		)
		return

	#--------------------------------------------------------------
	# regular invocation
	#--------------------------------------------------------------

	if args.xmlonly:
		args.html = False
		args.xml = True

	output_dir = Path.cwd()
	temp_dir = None
	if args.bug_report_internal:
		output_dir = bug_report_directory / r'output'
		temp_dir = bug_report_directory / r'temp'
		args.verbose = True
		args.nocleanup = True

	with ScopeTimer(r'All tasks', print_start=False, print_end=True) as timer:
		run(
			# named args:
			config_path=args.config,
			output_dir=output_dir,
			output_html=args.html,
			output_xml=args.xml,
			threads=args.threads,
			cleanup=not args.nocleanup,
			verbose=args.verbose,
			logger=True,  # stderr + stdout
			html_include=args.ppinclude,
			html_exclude=args.ppexclude,
			treat_warnings_as_errors=args.werror,
			theme=args.theme,
			copy_assets=not args.noassets,
			temp_dir=temp_dir,
			# kwargs:
			xml_v2=args.xml_v2
		)



def main_blog_post(invoker=True):
	"""
	The entry point when the library is invoked as `poxyblog`.
	"""
	if invoker:
		_invoker(main_blog_post, invoker=False)
		return

	args = argparse.ArgumentParser(
		description=r'Initializes a new blog post for Poxy sites.', formatter_class=argparse.RawTextHelpFormatter
	)
	args.add_argument(
		r'title',  #
		type=str,
		help=r'the title of the new blog post'
	)
	args.add_argument(
		r'-v',  #
		r'--verbose',
		action=r'store_true',
		help=r"enable very noisy diagnostic output"
	)
	args.add_argument(
		r'--version',  #
		action=r'store_true',
		help=r"print the version and exit",
		dest=r'print_version'
	)
	args = args.parse_args()

	if args.print_version:
		print(r'.'.join([str(v) for v in lib_version()]))
		return

	date = datetime.datetime.now().date()

	title = args.title.strip()
	if not title:
		raise Error(r'title cannot be blank.')
	if re.search(r''''[\n\v\f\r]''', title) is not None:
		raise Error(r'title cannot contain newline characters.')
	file = re.sub(r'''[!@#$%^&*;:'"<>?/\\\s|+]+''', '_', title)
	file = rf'{date}_{file.lower()}.md'

	blog_dir = Path(r'blog')
	if blog_dir.exists() and not blog_dir.is_dir():
		raise Error(rf'{blog_dir.resolve()} already exists and is not a directory')
	blog_dir.mkdir(exist_ok=True)

	file = Path(blog_dir, file)
	if file.exists():
		if not file.is_file():
			raise Error(rf'{file.resolve()} already exist and is not a file')
		raise Error(rf'{file.resolve()} already exists')

	with open(file, r'w', encoding=r'utf-8', newline='\n') as f:
		write = lambda s='', end='\n': print(s, end=end, file=f)
		write(rf'# {title}')
		write()
		write()
		write()
		write()
		write()
		write(r'<!--[poxy_metadata[')
		write(rf'tags = []')
		write(r']]-->')
	print(rf'Blog post file initialized: {file.resolve()}')



if __name__ == '__main__':
	main()



#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT
"""
The 'actually do the thing' module.
"""

import os
import subprocess
import concurrent.futures as futures
import tempfile
from io import StringIO
from lxml import etree
from .utils import *
from . import doxygen
from . import soup
from . import fixers
from . import graph
from . import xmlu
from .project import Context
from .svg import SVG
from distutils.dir_util import copy_tree
from trieregex import TrieRegEx

#=======================================================================================================================
# HELPERS
#=======================================================================================================================



def make_temp_file():
	return tempfile.SpooledTemporaryFile(mode='w+', newline='\n', encoding='utf-8')



#=======================================================================================================================
# PRE/POST PROCESSORS
#=======================================================================================================================

DOXYGEN_DEFAULTS = (
	(r'ALLEXTERNALS', False),
	(r'ALLOW_UNICODE_NAMES', False),
	(r'ALWAYS_DETAILED_SEC', False),
	(r'AUTOLINK_SUPPORT', True),
	(r'BUILTIN_STL_SUPPORT', False),
	(r'CASE_SENSE_NAMES', False),
	(r'CLASS_DIAGRAMS', False),
	(r'CPP_CLI_SUPPORT', False),
	(r'CREATE_SUBDIRS', False),
	(r'DISTRIBUTE_GROUP_DOC', False),
	(r'DOXYFILE_ENCODING', r'UTF-8'),
	(r'DOT_FONTNAME', r'Source Sans Pro'),
	(r'DOT_FONTSIZE', 16),
	(r'ENABLE_PREPROCESSING', True),
	(r'EXAMPLE_RECURSIVE', False),
	(r'EXCLUDE_SYMLINKS', False),
	(r'EXPAND_ONLY_PREDEF', False),
	(r'EXTERNAL_GROUPS', False),
	(r'EXTERNAL_PAGES', False),
	(r'EXTRACT_ANON_NSPACES', False),
	(r'EXTRACT_LOCAL_CLASSES', False),
	(r'EXTRACT_LOCAL_METHODS', False),
	(r'EXTRACT_PACKAGE', False),
	(r'EXTRACT_PRIV_VIRTUAL', True),
	(r'EXTRACT_PRIVATE', False),
	(r'EXTRACT_STATIC', False),
	(r'FILTER_PATTERNS', None),
	(r'FILTER_SOURCE_FILES', False),
	(r'FILTER_SOURCE_PATTERNS', None),
	(r'FORCE_LOCAL_INCLUDES', False),
	(r'FULL_PATH_NAMES', True),
	(r'GENERATE_AUTOGEN_DEF', False),
	(r'GENERATE_BUGLIST', False),
	(r'GENERATE_CHI', False),
	(r'GENERATE_DEPRECATEDLIST', False),
	(r'GENERATE_DOCBOOK', False),
	(r'GENERATE_DOCSET', False),
	(r'GENERATE_ECLIPSEHELP', False),
	(r'GENERATE_HTML', False),
	(r'GENERATE_HTMLHELP', False),
	(r'GENERATE_LATEX', False),
	(r'GENERATE_LEGEND', False),
	(r'GENERATE_MAN', False),
	(r'GENERATE_PERLMOD', False),
	(r'GENERATE_QHP', False),
	(r'GENERATE_RTF', False),
	(r'GENERATE_SQLITE3', False),
	(r'GENERATE_TESTLIST', False),
	(r'GENERATE_TODOLIST', False),
	(r'GENERATE_TREEVIEW', False),
	(r'GENERATE_XML', True),
	(r'HAVE_DOT', False),
	(r'HIDE_COMPOUND_REFERENCE', False),
	(r'HIDE_FRIEND_COMPOUNDS', False),
	(r'HIDE_IN_BODY_DOCS', False),
	(r'HIDE_SCOPE_NAMES', False),
	(r'HIDE_UNDOC_CLASSES', True),
	(r'HIDE_UNDOC_MEMBERS', True),
	(r'HTML_EXTRA_STYLESHEET', None),
	(r'HTML_FILE_EXTENSION', r'.html'),
	(r'HTML_OUTPUT', r'html'),
	(r'IDL_PROPERTY_SUPPORT', False),
	(r'INHERIT_DOCS', True),
	(r'INLINE_GROUPED_CLASSES', False),
	(r'INLINE_INFO', True),
	(r'INLINE_INHERITED_MEMB', True),
	(r'INLINE_SIMPLE_STRUCTS', False),
	(r'INLINE_SOURCES', False),
	(r'INPUT_ENCODING', r'UTF-8'),
	(r'INPUT_FILTER', None),
	(r'LOOKUP_CACHE_SIZE', 2),
	(r'MACRO_EXPANSION', True),
	(r'MARKDOWN_SUPPORT', True),
	(r'OPTIMIZE_FOR_FORTRAN', False),
	(r'OPTIMIZE_OUTPUT_FOR_C', False),
	(r'OPTIMIZE_OUTPUT_JAVA', False),
	(r'OPTIMIZE_OUTPUT_SLICE', False),
	(r'OPTIMIZE_OUTPUT_VHDL', False),
	(r'PYTHON_DOCSTRING', True),
	(r'QUIET', False),
	(r'RECURSIVE', False),
	(r'REFERENCES_LINK_SOURCE', False),
	(r'RESOLVE_UNNAMED_PARAMS', True),
	(r'SEARCH_INCLUDES', False),
	(r'SEPARATE_MEMBER_PAGES', False),
	(r'SHORT_NAMES', False),
	(r'SHOW_GROUPED_MEMB_INC', False),
	(r'SHOW_USED_FILES', False),
	(r'SIP_SUPPORT', False),
	(r'SKIP_FUNCTION_MACROS', False),
	(r'SORT_BRIEF_DOCS', False),
	(r'SORT_BY_SCOPE_NAME', False),
	(r'SORT_GROUP_NAMES', True),
	(r'SORT_MEMBER_DOCS', False),
	(r'SORT_MEMBERS_CTORS_1ST', True),
	(r'SOURCE_BROWSER', False),
	(r'STRICT_PROTO_MATCHING', False),
	(r'STRIP_FROM_INC_PATH', None),  # we handle this
	(r'SUBGROUPING', True),
	(r'TAB_SIZE', 4),
	(r'TOC_INCLUDE_HEADINGS', 3),
	(r'TYPEDEF_HIDES_STRUCT', False),
	(r'UML_LOOK', False),
	(r'USE_HTAGS', False),
	(r'USE_MDFILE_AS_MAINPAGE', None),
	(r'VERBATIM_HEADERS', False),
	(r'WARN_AS_ERROR', False),  # we handle this
	(r'WARN_IF_DOC_ERROR', True),
	(r'WARN_IF_INCOMPLETE_DOC', True),
	(r'WARN_LOGFILE', None),
	(r'XML_NS_MEMB_FILE_SCOPE', True),
	(r'XML_PROGRAMLISTING', False),
)



def preprocess_doxyfile(context: Context):
	assert context is not None
	assert isinstance(context, Context)

	with doxygen.Doxyfile(
		input_path=None, output_path=context.doxyfile_path, cwd=context.input_dir, logger=context.verbose_logger
	) as df:

		df.append()
		df.append(r'#---------------------------------------------------------------------------')
		df.append(r'# marzer/poxy')
		df.append(r'#---------------------------------------------------------------------------', end='\n\n')

		df.append(r'# doxygen defaults', end='\n\n')  # ----------------------------------------

		for k, v in DOXYGEN_DEFAULTS:
			df.set_value(k, v)

		df.append()
		df.append(r'# general config', end='\n\n')  # ---------------------------------------------------

		df.set_value(r'OUTPUT_DIRECTORY', context.output_dir)
		df.set_value(r'XML_OUTPUT', context.temp_xml_dir)
		df.set_value(r'PROJECT_NAME', context.name)
		df.set_value(r'PROJECT_BRIEF', context.description)
		df.set_value(r'PROJECT_LOGO', context.logo)
		df.set_value(r'SHOW_INCLUDE_FILES', context.show_includes)
		df.set_value(r'INTERNAL_DOCS', context.internal_docs)
		df.add_value(
			r'ENABLED_SECTIONS', (r'private', r'internal') if context.internal_docs else (r'public', r'external')
		)
		df.add_value(r'ENABLED_SECTIONS', r'poxy_supports_concepts')
		if context.xml_v2:
			df.set_value(r'INLINE_INHERITED_MEMB', False)

		if context.generate_tagfile:
			df.set_value(r'GENERATE_TAGFILE', context.tagfile_path)
		else:
			df.set_value(r'GENERATE_TAGFILE', None)

		df.set_value(r'NUM_PROC_THREADS', min(context.threads, 32))
		df.add_value(r'CLANG_OPTIONS', rf'-std=c++{context.cpp%100}')
		df.add_value(r'CLANG_OPTIONS', r'-Wno-everything')

		home_md_path = None
		for home_md in (r'HOME.md', r'home.md', r'INDEX.md', r'index.md', r'README.md', r'readme.md'):
			p = Path(context.input_dir, home_md)
			if p.exists() and p.is_file():
				home_md_path = p
				break
		if home_md_path is not None:
			home_md_temp_path = Path(context.temp_pages_dir, r'home.md')
			copy_file(home_md_path, home_md_temp_path, logger=context.verbose_logger)
			df.set_value(r'USE_MDFILE_AS_MAINPAGE', home_md_temp_path)

		df.append()
		df.append(r'# context.warnings', end='\n\n')  # ---------------------------------------------------

		df.set_value(r'WARNINGS', context.warnings.enabled)
		df.set_value(r'WARN_IF_UNDOCUMENTED', context.warnings.undocumented)

		df.append()
		df.append(r'# context.sources', end='\n\n')  # ----------------------------------------------------

		df.add_value(r'INPUT', context.sources.paths)
		df.set_value(r'FILE_PATTERNS', context.sources.patterns)
		df.add_value(r'EXCLUDE', context.html_dir)
		df.add_value(r'STRIP_FROM_PATH', context.sources.strip_paths)
		df.set_value(r'EXTRACT_ALL', context.sources.extract_all)

		df.append()
		df.append(r'# context.examples', end='\n\n')  # ----------------------------------------------------

		df.add_value(r'EXAMPLE_PATH', context.examples.paths)
		df.set_value(r'EXAMPLE_PATTERNS', context.examples.patterns)

		if context.images.paths:  # ----------------------------------------------------
			df.append()
			df.append(r'# context.images', end='\n\n')
			df.add_value(r'IMAGE_PATH', context.images.paths)

		if context.tagfiles:  # ----------------------------------------------------
			df.append()
			df.append(r'# context.tagfiles', end='\n\n')
			df.add_value(r'TAGFILES', [rf'{file}={dest}' for _, (file, dest) in context.tagfiles.items()])

		if context.aliases:  # ----------------------------------------------------
			df.append()
			df.append(r'# context.aliases', end='\n\n')
			df.add_value(r'ALIASES', [rf'{k}={v}' for k, v in context.aliases.items()])

		if context.macros:  # ----------------------------------------------------
			df.append()
			df.append(r'# context.macros', end='\n\n')
			df.add_value(r'PREDEFINED', [rf'{k}={v}' for k, v in context.macros.items()])

		df.cleanup()
		context.verbose(r'Doxyfile:')
		context.verbose(df.get_text(), indent=r'    ')



def preprocess_changelog(context: Context):
	assert context is not None
	assert isinstance(context, Context)
	if not context.changelog:
		return

	# make sure we're working with a temp copy, not the user's actual changelog
	# (the actual copying should already be done in the context's initialization)
	assert context.changelog.parent == context.temp_pages_dir
	assert_existing_file(context.changelog)

	text = read_all_text_from_file(context.changelog, logger=context.verbose_logger).strip()
	text = text.replace('\r\n', '\n')
	text = re.sub(r'\n<br[ \t]*/?><br[ \t]*/?>\n', r'', text)
	if context.repo:
		text = re.sub(r'#([0-9]+)', lambda m: rf'[#{m[1]}]({context.repo.make_issue_uri(m[1])})', text)
		text = re.sub(r'!([0-9]+)', lambda m: rf'[!{m[1]}]({context.repo.make_pull_request_uri(m[1])})', text)
		text = re.sub(r'@([a-zA-Z0-9_-]+)', lambda m: rf'[@{m[1]}]({context.repo.make_user_uri(m[1])})', text)
	text = text.replace(r'&amp;', r'__poxy_thiswasan_amp')
	text = text.replace(r'&#xFE0F;', r'__poxy_thiswasan_fe0f')
	text = text.replace(r'@', r'__poxy_thiswasan_at')
	text = f'\n{text}\n'
	text = re.sub('\n#[^#].+?\n', '\n', text)
	text = f'@page poxy_changelog Changelog\n\n@tableofcontents\n\n{text}'
	text = text.rstrip()
	text += '\n\n'
	context.verbose(rf'Writing {context.changelog}')
	with open(context.changelog, r'w', encoding=r'utf-8', newline='\n') as f:
		f.write(text)



def preprocess_tagfiles(context: Context):
	assert context is not None
	assert isinstance(context, Context)
	if not context.unresolved_tagfiles:
		return
	with ScopeTimer(r'Resolving remote tagfiles', print_start=True, print_end=context.verbose_logger) as t:
		for source, (file, _) in context.tagfiles.items():
			if file.exists() or not is_uri(source):
				continue
			context.verbose(rf'Downloading {source}')
			text = download_text(source, timeout=30)
			context.verbose(rf'Writing {file}')
			with open(file, 'w', encoding='utf-8', newline='\n') as f:
				f.write(text)



def postprocess_xml(context: Context):
	assert context is not None
	assert isinstance(context, Context)

	xml_files = get_all_files(context.temp_xml_dir, any=(r'*.xml'))
	if not xml_files:
		return

	context.verbose(rf'Post-processing {len(xml_files) + len(context.tagfiles)} XML files...')

	inline_namespace_ids = None
	if context.inline_namespaces:
		inline_namespace_ids = [f'namespace{doxygen.mangle_name(ns)}' for ns in context.inline_namespaces]

	implementation_header_data = None
	implementation_header_mappings = None
	implementation_header_innernamespaces = None
	implementation_header_sectiondefs = None
	implementation_header_unused_keys = None
	implementation_header_unused_values = None
	if context.implementation_headers:
		implementation_header_data = [(
			hp, os.path.basename(hp), doxygen.mangle_name(os.path.basename(hp)),
			[(i, os.path.basename(i), doxygen.mangle_name(os.path.basename(i))) for i in impl]
		) for hp, impl in context.implementation_headers]
		implementation_header_unused_keys = set()
		for hp, impl in context.implementation_headers:
			implementation_header_unused_keys.add(hp)
		implementation_header_unused_values = dict()
		for hdata in implementation_header_data:
			for (ip, ifn, iid) in hdata[3]:
				implementation_header_unused_values[iid] = (ip, hdata[0])
		implementation_header_mappings = dict()
		implementation_header_innernamespaces = dict()
		implementation_header_sectiondefs = dict()
		for hdata in implementation_header_data:
			implementation_header_innernamespaces[hdata[2]] = []
			implementation_header_sectiondefs[hdata[2]] = []
			for (ip, ifn, iid) in hdata[3]:
				implementation_header_mappings[iid] = hdata

	context.compounds = dict()
	context.compound_pages = dict()
	context.compound_kinds = set()

	# process xml files
	if 1:

		# pre-pass to delete junk files
		if 1:
			# 'file' entries for markdown and dox files
			dox_files = [rf'*{doxygen.mangle_name(ext)}.xml' for ext in (r'.dox', r'.md')]
			dox_files.append(r'md_home.xml')
			for xml_file in get_all_files(context.temp_xml_dir, any=dox_files):
				delete_file(xml_file, logger=context.verbose_logger)

			# 'dir' entries for empty directories
			deleted = True
			while deleted:
				deleted = False
				for xml_file in get_all_files(context.temp_xml_dir, all=(r'dir*.xml')):
					root = xmlu.read(xml_file)
					compounddef = root.find(r'compounddef')
					if compounddef is None or compounddef.get(r'kind') != r'dir':
						continue
					existing_inners = 0
					for subtype in (r'innerfile', r'innerdir'):
						for inner in compounddef.findall(subtype):
							ref_file = Path(context.temp_xml_dir, rf'{inner.get(r"refid")}.xml')
							if ref_file.exists():
								existing_inners = existing_inners + 1
					if not existing_inners:
						delete_file(xml_file, logger=context.verbose_logger)
						deleted = True

		extracted_implementation = False
		xml_files = [
			f for f in get_all_files(context.temp_xml_dir, any=(r'*.xml')) if f.name.lower() != r'doxyfile.xml'
		]
		all_inners_by_type = {r'namespace': set(), r'class': set(), r'concept': set()}

		# do '<doxygenindex>' first
		for xml_file in xml_files:

			root = xmlu.read(xml_file)
			if root.tag != r'doxygenindex':
				continue

			context.verbose(rf'Post-processing {xml_file}')
			changed = False

			# remove entries for files we might have explicitly deleted above
			for compound in [
				tag for tag in root.findall(r'compound') if tag.get(r'kind') in (r'file', r'dir', r'concept')
			]:
				ref_file = Path(context.temp_xml_dir, rf'{compound.get(r"refid")}.xml')
				if not ref_file.exists():
					root.remove(compound)
					changed = True

			# enumerate all compound pages and their types for later (e.g. HTML post-process)
			for tag in root.findall(r'compound'):
				refid = tag.get(r'refid').strip()
				assert refid
				filename = refid
				if refid == r'indexpage':
					filename = r'index'
				filename = filename + r'.html'
				context.compounds[refid] = {
					r'refid': refid,
					r'filename': filename,
					r'kind': tag.get(r'kind'),
					r'name': tag.find(r'name').text,
					r'title': tag.find(r'title').text.strip() if tag.find(r'title') is not None else r''
				}
				context.compound_pages[filename] = context.compounds[refid]
				context.compound_kinds.add(tag.get(r'kind'))

			if changed:
				xmlu.write(root, xml_file)

		# now do '<doxygen>' files
		for xml_file in xml_files:

			root = xmlu.read(xml_file)
			if root.tag != r'doxygen':
				continue

			context.verbose(rf'Post-processing {xml_file}')
			changed = False

			compounddef = root.find(r'compounddef')
			if compounddef is None:
				context.warning(rf'{xml_file} did not contain a <compounddef>!')
				continue

			compound_id = compounddef.get(r'id')
			if compound_id is None or not compound_id:
				context.warning(rf'{xml_file} did not have attribute "id"!')
				continue

			compound_kind = compounddef.get(r'kind')
			if compound_kind is None or not compound_kind:
				context.warning(rf'{xml_file} did not have attribute "kind"!')
				continue

			compound_name = compounddef.find(r'compoundname')
			if compound_name is None or not compound_name.text:
				context.warning(rf'{xml_file} did not contain a valid <compoundname>!')
				continue
			compound_name = str(compound_name.text).strip()

			compound_filename = rf'{compound_id}.html'
			if compound_id == r'indexpage':
				compound_filename = r'index.html'

			compound_title = compounddef.find(r'title')
			compound_title = compound_title.text if compound_title is not None else compound_name

			# add entry to compounds etc
			if compound_id not in context.compounds:
				context.compounds[compound_id] = {
					r'refid': compound_id,
					r'filename': compound_filename,
					r'kind': compound_kind,
					r'name': compound_name,
					r'title': compound_title
				}
				context.compound_pages[compound_filename] = context.compounds[compound_id]
			compound_page = context.compound_pages[compound_filename]
			if r'title' not in compound_page or not compound_page[r'title']:
				compound_page[r'title'] = compound_title

			if compound_kind != r'page':

				# merge user-defined sections with the same header name
				sectiondefs = [s for s in compounddef.findall(r'sectiondef') if s.get(r'kind') == r'user-defined']
				sections_with_headers = dict()
				for section in sectiondefs:
					header = section.find(r'header')
					if header is not None and header.text:
						if header.text not in sections_with_headers:
							sections_with_headers[header.text] = []
						sections_with_headers[header.text].append(section)
				for key, vals in sections_with_headers.items():
					if len(vals) > 1:
						first_section = vals.pop(0)
						for section in vals:
							for member in section.findall(r'memberdef'):
								section.remove(member)
								first_section.append(member)
							compounddef.remove(section)
							changed = True

				# sort user-defined sections based on their name
				sectiondefs = [s for s in compounddef.findall(r'sectiondef') if s.get(r'kind') == r'user-defined']
				sectiondefs = [s for s in sectiondefs if s.find(r'header') is not None]
				for section in sectiondefs:
					compounddef.remove(section)
				sectiondefs.sort(key=lambda s: s.find(r'header').text)
				for section in sectiondefs:
					compounddef.append(section)
					changed = True

				# per-section stuff
				for section in compounddef.findall(r'sectiondef'):

					# remove members which are listed multiple times because doxygen is idiotic:
					members = [tag for tag in section.findall(r'memberdef')]
					for i in range(len(members) - 1, 0, -1):
						for j in range(i):
							if members[i].get(r'id') == members[j].get(r'id'):
								section.remove(members[i])
								changed = True
								break

					# fix keywords like 'friend' erroneously included in the type
					if 1:

						members = [
							m for m in section.findall(r'memberdef')
							if m.get(r'kind') in (r'friend', r'function', r'variable')
						]

						# leaked keywords
						attribute_keywords = (
							(r'constexpr', r'constexpr', r'yes'),  #
							(r'constinit', r'constinit', r'yes'),
							(r'consteval', r'consteval', r'yes'),
							(r'explicit', r'explicit', r'yes'),
							(r'static', r'static', r'yes'),
							(r'friend', None, None),
							(r'extern', None, None),
							(r'inline', r'inline', r'yes'),
							(r'virtual', r'virt', r'virtual')
						)
						for member in members:
							type = member.find(r'type')
							if type is None or type.text is None:
								continue
							matched_bad_keyword = True
							while matched_bad_keyword:
								matched_bad_keyword = False
								for kw, attr, attr_value in attribute_keywords:
									if type.text == kw:  # constructors
										type.text = ''
									elif type.text.startswith(kw + ' '):
										type.text = type.text[len(kw):].strip()
									elif type.text.endswith(' ' + kw):
										type.text = type.text[:len(kw)].strip()
									else:
										continue
									matched_bad_keyword = True
									changed = True
									if attr is not None:
										member.set(attr, attr_value)
									elif kw == r'friend':
										member.set(r'kind', r'friend')

					# fix goofy parsing of trailing return types
					if 1:

						members = [
							m for m in section.findall(r'memberdef') if m.get(r'kind') in (r'friend', r'function')
						]

						# trailing return type bug (https://github.com/mosra/m.css/issues/94)
						for member in members:
							type_elem = member.find(r'type')
							if type_elem is None or type_elem.text != r'auto':
								continue
							args_elem = member.find(r'argsstring')
							if args_elem is None or not args_elem.text or args_elem.text.find(r'decltype') != -1:
								continue
							match = re.search(r'^(.*?)\s*->\s*([a-zA-Z][a-zA-Z0-9_::*&<>\s]+?)\s*$', args_elem.text)
							if match:
								args_elem.text = str(match[1])
								trailing_return_type = str(match[2]).strip()
								trailing_return_type = re.sub(r'\s+', r' ', trailing_return_type)
								trailing_return_type = re.sub(r'(::|[<>*&])\s+', r'\1', trailing_return_type)
								trailing_return_type = re.sub(r'\s+(::|[<>*&])', r'\1', trailing_return_type)
								type_elem.text = trailing_return_type
								changed = True

					# re-sort members to override Doxygen's weird and stupid sorting 'rules'
					if 1:
						sort_members_by_name = lambda tag: tag.find(r'name').text
						members = [tag for tag in section.findall(r'memberdef')]
						for tag in members:
							section.remove(tag)
						# fmt: off
						# yapf: disable
						groups = [
							([tag for tag in members if tag.get(r'kind') == r'define'], True),  #
							([tag for tag in members if tag.get(r'kind') == r'typedef'], True),
							([tag for tag in members if tag.get(r'kind') == r'concept'], True),
							([tag for tag in members if tag.get(r'kind') == r'enum'], True),
							([tag for tag in members if tag.get(r'kind') == r'variable' and tag.get(r'static') == r'yes'], True),
							([tag for tag in members if tag.get(r'kind') == r'variable' and tag.get(r'static') == r'no'], compound_kind not in (r'class', r'struct', r'union')),
							([tag for tag in members if tag.get(r'kind') == r'function' and tag.get(r'static') == r'yes'], True),
							([tag for tag in members if tag.get(r'kind') == r'function' and tag.get(r'static') == r'no'], True),
							([tag for tag in members if tag.get(r'kind') == r'friend'], True)
						]
						# yapf: enable
						# fmt: on
						for group, sort in groups:
							if sort:
								group.sort(key=sort_members_by_name)
							for tag in group:
								members.remove(tag)
								section.append(tag)
								changed = True
						# if we've missed any groups just glob them on the end
						if members:
							members.sort(key=sort_members_by_name)
							changed = True
							for tag in members:
								section.append(tag)

			# namespaces
			if compound_kind == r'namespace':

				# set inline namespaces
				if context.inline_namespaces:
					for nsid in inline_namespace_ids:
						if compound_id == nsid:
							compounddef.set(r'inline', r'yes')
							changed = True
							break

			# dirs
			if compound_kind == r'dir':

				# remove implementation headers
				if context.implementation_headers:
					for innerfile in compounddef.findall(r'innerfile'):
						if innerfile.get(r'refid') in implementation_header_mappings:
							compounddef.remove(innerfile)
							changed = True

			# files
			if compound_kind == r'file':

				# simplify the XML by removing unnecessary junk
				for tag in (r'includes', r'includedby', r'incdepgraph', r'invincdepgraph'):
					for t in compounddef.findall(tag):
						compounddef.remove(t)
						changed = True

				# rip the good bits out of implementation headers
				if context.implementation_headers:
					iid = compound_id
					if iid in implementation_header_mappings:
						hid = implementation_header_mappings[iid][2]
						innernamespaces = compounddef.findall(r'innernamespace')
						if innernamespaces:
							implementation_header_innernamespaces[
								hid] = implementation_header_innernamespaces[hid] + innernamespaces
							extracted_implementation = True
							if iid in implementation_header_unused_values:
								del implementation_header_unused_values[iid]
							for tag in innernamespaces:
								compounddef.remove(tag)
								changed = True
						sectiondefs = compounddef.findall(r'sectiondef')
						if sectiondefs:
							implementation_header_sectiondefs[hid
																] = implementation_header_sectiondefs[hid] + sectiondefs
							extracted_implementation = True
							if iid in implementation_header_unused_values:
								del implementation_header_unused_values[iid]
							for tag in sectiondefs:
								compounddef.remove(tag)
								changed = True

			# groups and namespaces
			if compound_kind in (r'group', r'namespace'):

				# fix inner(class|namespace|group|concept) sorting
				inners = [tag for tag in compounddef.iterchildren() if tag.tag.startswith(r'inner')]
				if inners:
					changed = True
					for tag in inners:
						compounddef.remove(tag)
					inners.sort(key=lambda tag: tag.text)
					for tag in inners:
						compounddef.append(tag)

			# all namespace 'innerXXXXXX'
			if compound_kind in (r'namespace', r'struct', r'class', r'union', r'concept'):
				if compound_name.rfind(r'::') != -1:
					all_inners_by_type[r'class' if compound_kind in (r'struct', r'union') else compound_kind].add(
						(compound_id, compound_name)
					)

			if changed:
				xmlu.write(root, xml_file)

		context.verbose_value(r'Context.compounds', context.compounds)
		context.verbose_value(r'Context.compound_pages', context.compound_pages)
		context.verbose_value(r'Context.compound_kinds', context.compound_kinds)

		# fix up namespaces/classes that are missing <innerXXXX> nodes
		if 1:
			outer_namespaces = dict()
			for inner_type, ids_and_names in all_inners_by_type.items():
				for id, name in ids_and_names:
					ns = name[:name.rfind(r'::')]
					assert ns
					if ns not in outer_namespaces:
						outer_namespaces[ns] = []
					outer_namespaces[ns].append((inner_type, id, name))
			for ns, vals in outer_namespaces.items():
				xml_file = None
				for outer_type in (r'namespace', r'struct', r'class', r'union'):
					f = Path(context.temp_xml_dir, rf'{outer_type}{doxygen.mangle_name(ns)}.xml')
					if f.exists():
						xml_file = f
						break
				if not xml_file:
					continue
				root = xmlu.read(xml_file)
				compounddef = root.find(r'compounddef')
				if compounddef is None:
					continue
				changed = False
				existing_inner_ids = set()
				for inner_type in (r'class', r'namespace', r'concept'):
					for elem in compounddef.findall(rf'inner{inner_type}'):
						id = elem.get(r'refid')
						if id:
							existing_inner_ids.add(str(id))
				for (inner_type, id, name) in vals:
					if id not in existing_inner_ids:
						elem = xmlu.make_child(compounddef, rf'inner{inner_type}')
						elem.text = name
						elem.set(r'refid', id)
						elem.set(r'prot', r'public')  # todo: this isn't necessarily correct
						existing_inner_ids.add(id)
						changed = True
				if changed:
					xmlu.write(root, xml_file)

		# merge extracted implementations
		if extracted_implementation:
			for (hp, hfn, hid, impl) in implementation_header_data:
				xml_file = Path(context.temp_xml_dir, rf'{hid}.xml')
				context.verbose(rf'Merging implementation nodes into {xml_file}')
				root = xmlu.read(xml_file)
				compounddef = root.find(r'compounddef')
				changed = False

				innernamespaces = compounddef.findall(r'innernamespace')
				for new_tag in implementation_header_innernamespaces[hid]:
					matched = False
					for existing_tag in innernamespaces:
						if existing_tag.get(r'refid') == new_tag.get(r'refid'):
							matched = True
							break
					if not matched:
						compounddef.append(new_tag)
						innernamespaces.append(new_tag)
						changed = True

				sectiondefs = compounddef.findall(r'sectiondef')
				for new_section in implementation_header_sectiondefs[hid]:
					matched_section = False
					for existing_section in sectiondefs:
						if existing_section.get(r'kind') == new_section.get(r'kind'):
							matched_section = True

							memberdefs = existing_section.findall(r'memberdef')
							new_memberdefs = new_section.findall(r'memberdef')
							for new_memberdef in new_memberdefs:
								matched = False
								for existing_memberdef in memberdefs:
									if existing_memberdef.get(r'id') == new_memberdef.get(r'id'):
										matched = True
										break

								if not matched:
									new_section.remove(new_memberdef)
									existing_section.append(new_memberdef)
									memberdefs.append(new_memberdef)
									changed = True
							break

					if not matched_section:
						compounddef.append(new_section)
						sectiondefs.append(new_section)
						changed = True

				if changed:
					implementation_header_unused_keys.remove(hp)
					xmlu.write(root, xml_file)

		# sanity-check implementation header state
		if implementation_header_unused_keys:
			for key in implementation_header_unused_keys:
				context.warning(rf"implementation_header: nothing extracted for '{key}'")
		if implementation_header_unused_values:
			for iid, idata in implementation_header_unused_values.items():
				context.warning(rf"implementation_header: nothing extracted from '{idata[0]}' for '{idata[1]}'")

	# delete the impl header xml files
	if 1 and context.implementation_headers:
		for hdata in implementation_header_data:
			for (ip, ifn, iid) in hdata[3]:
				delete_file(Path(context.temp_xml_dir, rf'{iid}.xml'), logger=context.verbose_logger)

	# scan through the files and substitute impl header ids and paths as appropriate
	if 1 and context.implementation_headers:
		xml_files = get_all_files(context.temp_xml_dir, any=('*.xml'))
		for xml_file in xml_files:
			context.verbose(rf"Re-linking implementation headers in '{xml_file}'")
			xml_text = read_all_text_from_file(xml_file, logger=context.verbose_logger)
			for (hp, hfn, hid, impl) in implementation_header_data:
				for (ip, ifn, iid) in impl:
					#xml_text = xml_text.replace(f'refid="{iid}"',f'refid="{hid}"')
					xml_text = xml_text.replace(rf'compoundref="{iid}"', f'compoundref="{hid}"')
					xml_text = xml_text.replace(ip, hp)
			xmlu.write(xml_text, xml_file)



def postprocess_xml_v2(context: Context):
	assert context is not None
	assert isinstance(context, Context)

	log_func = lambda m: context.verbose(m)

	g = doxygen.read_graph_from_xml(context.temp_xml_dir, log_func=log_func)

	# delete 'file' nodes for markdown and dox files
	g.remove(filter=lambda n: n.type is graph.File and re.search(r'[.](?:md|dox)$', n.local_name, flags=re.I))

	# delete empty 'dir' nodes
	g.remove(filter=lambda n: n.type is graph.Directory and not len(list(n(graph.File, graph.Directory))))

	# todo:
	# - extract namespaces, types and enum values for syntax highlighting
	# - enumerate all compound pages and their types for later (e.g. HTML post-process)
	# - merge user-defined sections with the same name
	# - sort user-defined sections based on their name
	# - implementation headers

	for f in enumerate_files(context.temp_xml_dir, any=r'*.xml'):
		delete_file(f, logger=log_func)
	doxygen.write_graph_to_xml(g, context.temp_xml_dir, log_func=log_func)



def parse_xml(context: Context):
	assert context is not None
	assert isinstance(context, Context)

	xml_files = get_all_files(context.temp_xml_dir, any=(r'*.xml'))
	xml_files += [coerce_path(f) for _, (f, _) in context.tagfiles.items()]
	if context.generate_tagfile and context.tagfile_path:
		xml_files.append(context.tagfile_path)
	if not xml_files:
		return

	class Trie(object):

		def __init__(self):
			self.__trie = TrieRegEx()
			self.__count = 0

		def add(self, s: str):
			if not s:
				return
			self.__trie.add(s)
			self.__count += 1

		def __bool__(self) -> bool:
			return self.__count > 0

		def __str__(self) -> str:
			return self.__trie.regex()

	class Tries(object):

		def __init__(self):
			self.namespaces = Trie()
			self.types = Trie()
			self.enum_values = Trie()
			self.macros = Trie()
			self.functions = Trie()

	tries = Tries()

	def name_ok(s: str) -> bool:
		return s is not None and s and not re.search(r'[^a-zA-Z0-9_:]', s)

	def extract_all_members_from_compound_node(compound):
		nonlocal tries
		compound_name = compound.find(r'name')
		if compound_name is None or not name_ok(compound_name.text):
			return
		compound_kind = compound.get(r'kind')
		if compound_kind is None or compound_kind not in (
			r'namespace', r'class', r'struct', r'union', r'concept', r'group', r'file'
		):
			return
		# for files and groups we can only extract #defines because they need the full::namespace::context
		# otherwise we get all the C++ types
		member_kinds = (
			r'namespace', r'class', r'struct', r'union', r'concept', r'typedef', r'enum', r'enumvalue', r'function'
		)
		if compound_kind in (r'group', r'file'):
			member_kinds = (r'define', )
		members = [(m, m.find(r'name')) for m in compound.findall(r'member') if m.get(r'kind') in member_kinds]
		members = [(m, n) for m, n in members if n is not None and name_ok(n.text)]
		# first we do everything _except_ enumvalues because they require special handling
		enums = dict()
		for member, member_name in members:
			member_kind = member.get(r'kind')
			if member_kind == r'define':
				tries.macros.add(compound_name.text)
			else:
				member_qualified_name = rf'{compound_name.text}::{member_name.text}'
				if member_kind == r'namespace':
					tries.namespaces.add(member_qualified_name)
				elif member_kind == r'function':
					if member_name.text.startswith(r'operator'):
						continue
					tries.functions.add(member_qualified_name)
				elif member_kind != r'enumvalue':
					tries.types.add(member_qualified_name)
					if member_kind == r'enum':
						refid = member.get(r'refid')
						if refid:
							enums[refid] = member_qualified_name
		# then we do enumvaleus
		for member, member_name in members:
			if member.get(r'kind') != r'enumvalue':
				continue
			refid = member.get(r'refid')
			if not refid:
				continue
			for enum_refid, enum_qualified_name in enums.items():
				if refid.startswith(enum_refid):
					tries.enum_values.add(rf'{enum_qualified_name}::{member_name.text}')

	for xml_file in xml_files:
		if xml_file.name == r'Doxyfile.xml' or not xml_file.exists() or not xml_file.is_file():
			continue

		root = xmlu.read(xml_file)
		if root.tag not in (r'doxygenindex', r'tagfile'):
			continue

		context.verbose(rf'Extracting type information from {xml_file}')

		# tag files
		if root.tag == r'tagfile':

			def extract_types_from_tagfile_node(node):
				nonlocal tries
				namespaces = [(ns, ns.find(r'name')) for ns in node.findall(r'namespace')]
				namespaces = [(ns, n) for ns, n in namespaces if n is not None and name_ok(n.text)]
				for namespace, n in namespaces:
					tries.namespaces.add(n.text)

				classes = [(c, c.find(r'name')) for c in node.findall(r'class')
					if c.get(r'kind') in (r'class', r'struct', r'union')]
				classes = [(c, n) for c, n in classes if n is not None and name_ok(n.text)]
				for class_, n in classes:
					tries.types.add(n.text)

				compounds = [(c, c.find(r'name')) for c in node.findall(r'compound')
					if c.get(r'kind') in (r'namespace', r'class', r'struct', r'union', r'concept')]
				compounds = [(c, n) for c, n in compounds if n is not None and name_ok(n.text)]
				for compound, n in compounds:
					if compound.get(r'kind') == r'namespace':
						tries.namespaces.add(n.text)
					else:
						tries.types.add(n.text)
					extract_types_from_tagfile_node(compound)
					extract_all_members_from_compound_node(compound)

			extract_types_from_tagfile_node(root)

		# the doxygen index
		elif root.tag == r'doxygenindex':
			compounds = [(c, c.find(r'name')) for c in root.findall(r'compound')
				if c.get(r'kind') in (r'namespace', r'class', r'struct', r'union', r'concept')]
			compounds = [(c, n) for c, n in compounds if n is not None and name_ok(n.text)]
			for compound, n in compounds:
				if compound.get(r'kind') == r'namespace':
					tries.namespaces.add(n.text)
				else:
					tries.types.add(n.text)
				extract_all_members_from_compound_node(compound)

	# add to syntax highlighter
	if tries.namespaces:
		context.code_blocks.namespaces.add(str(tries.namespaces))
	if tries.types:
		context.code_blocks.types.add(str(tries.types))
	if tries.enum_values:
		context.code_blocks.enums.add(str(tries.enum_values))
	if tries.macros:
		context.code_blocks.macros.add(str(tries.macros))
	if tries.functions:
		context.code_blocks.functions.add(str(tries.functions))
	context.verbose_object(r'Context.code_blocks', context.code_blocks)



def clean_xml(context: Context):
	assert context is not None
	assert isinstance(context, Context)

	xml_files = get_all_files(context.temp_xml_dir, any=(r'*.xml'))
	for xml_file in xml_files:
		root = xmlu.read(
			xml_file,  #
			parser=xmlu.create_parser(remove_blank_text=True),
			logger=context.verbose_logger
		)

		# some description nodes end up with just whitespace; I guess lxml gets a bit confused here
		for elem in root.iter(r'briefdescription', r'detaileddescription', r'inbodydescription'):
			if len(elem) or not elem.text:
				continue
			if elem.text.strip() == r'':
				elem.text = r''

		# indent() will fuck up the formatting of some 'document-style' elements so we need to find and
		# extract those elements before prettifying the overall document
		sacred_elements = []
		sacred_element_ids = set()
		for elem in root.iter(r'programlisting', r'initializer', r'formula', r'computeroutput'):
			already_processed = False
			for parent in elem.iterancestors():
				if id(parent) in sacred_element_ids:
					already_processed = True
					break
			if already_processed:
				continue
			parent = elem.getparent()
			sacred_elements.append((elem, parent, parent.index(elem)))
			sacred_element_ids.add(id(elem))
		for elem, parent, _ in reversed(sacred_elements):
			parent.remove(elem)

		etree.indent(root, space='\t')

		# re-insert the extracted elements in their original positions
		for elem, parent, index in sacred_elements:
			parent.insert(index, elem)

		xmlu.write(
			root,  #
			xml_file,
			logger=context.verbose_logger
		)



def compile_regexes(context: Context):
	assert context is not None
	assert isinstance(context, Context)

	context.code_blocks.namespaces = regex_or(
		context.code_blocks.namespaces, pattern_prefix=r'(?:::)?', pattern_suffix=r'(?:::)?'
	)
	context.code_blocks.types = regex_or(
		context.code_blocks.types, pattern_prefix=r'(?:::)?', pattern_suffix=r'(?:::)?'
	)
	context.code_blocks.enums = regex_or(context.code_blocks.enums, pattern_prefix=r'(?:::)?')
	context.code_blocks.functions = regex_or(context.code_blocks.functions, pattern_prefix=r'(?:::)?')
	context.code_blocks.macros = regex_or(context.code_blocks.macros)
	context.autolinks = tuple([(re.compile(r'(?<![a-zA-Z_])' + expr + r'(?![a-zA-Z_])'), uri)
		for expr, uri in context.autolinks])



def preprocess_mcss_config(context: Context):
	assert context is not None
	assert isinstance(context, Context)

	# build HTML_HEADER
	html_header = ''
	if 1:
		# stylesheets
		for stylesheet in context.stylesheets:
			html_header += f'<link href="{stylesheet}" rel="stylesheet" referrerpolicy="no-referrer" />\n'
		# scripts
		for script in context.scripts:
			html_header += f'<script src="{script}"></script>\n'
		if context.theme != r'custom':
			html_header += f'<script>initialize_theme("{context.theme}");</script>\n'
		# metadata
		def add_meta_kvp(key_name, key, content):
			nonlocal html_header
			html_header += f'<meta {key_name}="{key}" content="{content}">\n'

		add_meta = lambda key, content: add_meta_kvp(r'name', key, content)
		add_property = lambda key, content: add_meta_kvp(r'property', key, content)
		add_itemprop = lambda key, content: add_meta_kvp(r'itemprop', key, content)
		# metadata - project name
		if context.name:
			if r'twitter:title' not in context.meta_tags:
				add_meta(r'twitter:title', context.name)
			add_property(r'og:title', context.name)
			add_itemprop(r'name', context.name)
		# metadata - project author
		if context.author:
			if r'author' not in context.meta_tags:
				add_meta(r'author', context.author)
			add_property(r'article:author', context.author)
		# metadata - project description
		if context.description:
			if r'description' not in context.meta_tags:
				add_meta(r'description', context.description)
			if r'twitter:description' not in context.meta_tags:
				add_meta(r'twitter:description', context.description)
			add_property(r'og:description', context.description)
			add_itemprop(r'description', context.description)
		# metadata - robots
		if not context.robots:
			if r'robots' not in context.meta_tags:
				add_meta(r'robots', r'noindex, nofollow')
			if r'googlebot' not in context.meta_tags:
				add_meta(r'googlebot', r'noindex, nofollow')
		# metadata - misc
		if r'format-detection' not in context.meta_tags:
			add_meta(r'format-detection', r'telephone=no')
		if r'generator' not in context.meta_tags:
			add_meta(r'generator', rf'Poxy v{context.version_string}')
		if r'referrer' not in context.meta_tags:
			add_meta(r'referrer', r'strict-origin-when-cross-origin')
		# metadata - additional user-specified tags
		for name, content in context.meta_tags.items():
			add_meta(name, content)
		# html_header
		if context.html_header:
			html_header += f'{context.html_header}\n'
		html_header = html_header.rstrip()

	# build + write conf.py
	with StringIO(newline='\n') as conf_py:
		conf = lambda s='', end='\n': print(reindent(s, indent=''), file=conf_py, end=end)

		# basic properties
		conf(rf"DOXYFILE = r'{context.doxyfile_path}'")
		conf(r"STYLESHEETS = []")  # suppress the default behaviour
		conf(rf'HTML_HEADER = """{html_header}"""')
		if context.theme == r'dark':
			conf(r"THEME_COLOR = '#22272e'")
		elif context.theme == r'light':
			conf(r"THEME_COLOR = '#cb4b16'")
		if context.favicon:
			conf(rf"FAVICON = r'{context.favicon}'")
		elif context.theme == r'dark':
			conf(rf"FAVICON = 'favicon-dark.png'")
		elif context.theme == r'light':
			conf(rf"FAVICON = 'favicon-light.png'")
		conf(rf'SHOW_UNDOCUMENTED = {context.sources.extract_all}')
		conf(r'CLASS_INDEX_EXPAND_LEVELS = 3')
		conf(r'FILE_INDEX_EXPAND_LEVELS = 3')
		conf(r'CLASS_INDEX_EXPAND_INNER = True')
		conf(r'SEARCH_DOWNLOAD_BINARY = False')
		conf(r'SEARCH_DISABLED = False')

		# navbar
		NAVBAR_ALIASES = {
			# poxy -> doxygen
			r'classes': r'annotated',
			r'groups': r'modules'
		}
		NAVBAR_TO_KIND = {
			r'annotated': (r'class', r'struct', r'union'),
			r'concepts': (r'concept', ),
			r'namespaces': (r'namespace', ),
			r'pages': (r'page', ),
			r'modules': (r'group', ),
			r'files': (r'file', r'dir')
		}
		navbar = ([], [])
		if context.navbar:
			# populate the navbar
			bar = [(NAVBAR_ALIASES[b] if b in NAVBAR_ALIASES else b) for b in context.navbar]
			# remove links to index pages that will have no entries
			for i in range(len(bar)):
				if bar[i] not in NAVBAR_TO_KIND:
					continue
				found = False
				for kind in NAVBAR_TO_KIND[bar[i]]:
					if kind in context.compound_kinds:
						found = True
						break
				if not found:
					bar[i] = None
			bar = [b for b in bar if b is not None]
			# handle theme and repo links
			for i in range(len(bar)):
				bar[i] = bar[i].strip()
				if bar[i] == r'repo':
					if not context.repo:
						bar[i] = None
						continue
					icon_path = Path(dirs.DATA, context.repo.icon_filename)
					if icon_path.exists():
						svg = SVG(icon_path, logger=context.verbose_logger, root_id=r'poxy-repo-icon')
						bar[i] = (
							rf'<a title="View on {type(context.repo).__name__}" '
							+ rf'target="_blank" href="{context.repo.uri}" '
							+ rf'class="poxy-icon repo {context.repo.KEY}">{svg}</a>', []
						)
					else:
						bar[i] = None
				elif bar[i] == r'theme':
					svg = SVG(
						Path(dirs.DATA, r'poxy-icon-theme.svg'),
						logger=context.verbose_logger,
						root_id=r'poxy-theme-switch-img'
					)
					bar[i] = (
						r'<a title="Toggle dark and light themes" '
						+ r'id="poxy-theme-switch" href="javascript:void(null);" role="button" '
						+ rf'class="poxy-icon theme" onClick="toggle_theme(); return false;">{svg}</a>', []
					)
				elif bar[i] in context.compounds:
					bar[i] = (
						rf'<a href="{context.compounds[bar[i]]["filename"]}">{context.compounds[bar[i]]["title"]}</a>',
						[]
					)
				elif re.search(r'^\s*<\s*[aA]\s+', bar[i]):
					bar[i] = (bar[i], [])
				elif re.search(r'[.]html?\s*$', bar[i], flags=re.I) and not is_uri(bar[i]):
					if bar[i] in context.compound_pages:
						bar[i] = (rf'<a href="{bar[i]}">{context.compound_pages[bar[i]]["title"]}</a>', [])
					else:
						bar[i] = (rf'<a href="{bar[i]}">{bar[i]}</a>', [])
			bar = [b for b in bar if b is not None]
			# automatically overflow onto the second row
			split = min(max(int(len(bar) / 2) + len(bar) % 2, 2), len(bar))
			for b, i in ((bar[:split], 0), (bar[split:], 1)):
				for j in range(len(b)):
					if isinstance(b[j], tuple):
						navbar[i].append(b[j])
					else:
						navbar[i].append((None, b[j], []))
		for i in (0, 1):
			if navbar[i]:
				conf(f'LINKS_NAVBAR{i+1} = [\n\t', end='')
				conf(',\n\t'.join([rf'{b}' for b in navbar[i]]))
				conf(r']')
			else:
				conf(rf'LINKS_NAVBAR{i+1} = []')

		# footer
		conf(r"FINE_PRINT = r'''")
		footer = []
		if context.repo:
			footer.append(rf'<a href="{context.repo.uri}" target="_blank">{type(context.repo).__name__}</a>')
			footer.append(rf'<a href="{context.repo.issues_uri}" target="_blank">Report an issue</a>')
		if context.changelog:
			footer.append(rf'<a href="md_poxy_changelog.html">Changelog</a>')
		if context.license and context.license[r'uri']:
			footer.append(rf'<a href="{context.license["uri"]}" target="_blank">License</a>')
		if context.generate_tagfile:
			footer.append(
				rf'<a href="{context.tagfile_path.name}" target="_blank" type="text/xml" download>Doxygen tagfile</a>'
			)
		if footer:
			for i in range(1, len(footer)):
				footer[i] = r' &bull; ' + footer[i]
			footer.append(r'<br><br>')
		footer.append(r'Site generated using <a href="https://github.com/marzer/poxy/">Poxy</a>')
		for i in range(len(footer)):
			conf(rf"    {footer[i]}")
		conf(r"'''")

		conf_py_text = conf_py.getvalue()
		context.verbose(r'm.css conf.py:')
		context.verbose(conf_py_text, indent=r'   ')

		# write conf.py
		context.verbose(rf'Writing {context.mcss_conf_path}')
		with open(context.mcss_conf_path, r'w', encoding=r'utf-8', newline='\n') as f:
			f.write(conf_py_text)



_worker_context = None



def _initialize_worker(context):
	global _worker_context
	_worker_context = context



def postprocess_html_file(path, context: Context = None):
	assert path is not None
	assert isinstance(path, Path)
	assert path.is_absolute()
	assert path.exists()

	if context is None:
		global _worker_context
		context = _worker_context
	assert context is not None
	assert isinstance(context, Context)

	context.info(rf'Post-processing {path}')
	text = None
	html = None

	def switch_to_html():
		nonlocal context
		nonlocal text
		nonlocal html
		if html is not None:
			return
		html = soup.HTMLDocument(text, logger=context.verbose_logger)

	def switch_to_text():
		nonlocal context
		nonlocal text
		nonlocal html
		if html is None:
			return
		html.smooth()
		text = str(html)
		html = None

	try:
		text = read_all_text_from_file(path, logger=context.verbose_logger)
		changed = False

		for fix in context.fixers:
			if isinstance(fix, fixers.HTMLFixer):
				switch_to_html()
				if fix(context, html, path):
					changed = True
					html.smooth()
			elif isinstance(fix, fixers.PlainTextFixer):
				switch_to_text()
				prev_text = text
				text = fix(context, prev_text, path)
				changed = changed or prev_text != text

		if changed:
			switch_to_text()
			context.verbose(rf'Writing {path}')
			with open(path, 'w', encoding='utf-8', newline='\n') as f:
				f.write(text)

	except Exception as e:
		context.info(rf'{type(e).__name__} raised while post-processing {path}')
		raise
	except:
		context.info(rf'Error occurred while post-processing {path}')
		raise



def postprocess_html(context: Context):
	assert context is not None
	assert isinstance(context, Context)

	files = filter_filenames(
		get_all_files(context.html_dir, any=('*.html', '*.htm')), context.html_include, context.html_exclude
	)
	if not files:
		return

	context.fixers = (
		fixers.MarkTOC(),
		fixers.Pygments(),
		fixers.CodeBlocks(),
		fixers.Banner(),
		fixers.CPPModifiers1(),
		fixers.CPPModifiers2(),
		fixers.StripIncludes(),
		fixers.AutoDocLinks(),
		fixers.Links(),
		fixers.CustomTags(),
		fixers.EmptyTags(),
		fixers.ImplementationDetails(),
		fixers.MarkdownPages(),
		fixers.InjectSVGs(),
		fixers.InstallSearchShim(),
	)

	threads = min(len(files), context.threads, 16)
	context.info(rf'Post-processing {len(files)} HTML files on {threads} thread{"s" if threads > 1 else ""}...')
	if threads > 1:
		with futures.ProcessPoolExecutor(
			max_workers=threads, initializer=_initialize_worker, initargs=(context, )
		) as executor:
			jobs = [executor.submit(postprocess_html_file, file) for file in files]
			for future in futures.as_completed(jobs):
				try:
					future.result()
				except:
					try:
						executor.shutdown(wait=False, cancel_futures=True)
					except TypeError:
						executor.shutdown(wait=False)
					raise

	else:
		for file in files:
			postprocess_html_file(file, context)



#=======================================================================================================================
# RUN
#=======================================================================================================================



def read_output_streams(stdout, stderr):
	stdout.seek(0)
	stderr.seek(0)
	return {r'stdout': stdout.read().strip(), r'stderr': stderr.read().strip()}



def dump_output_streams(context, outputs, source=''):
	if source:
		source = rf'{source} '
	if outputs[r'stdout']:
		context.info(rf'{source}stdout:')
		context.info(outputs[r'stdout'], indent=r'    ')
	if outputs[r'stderr']:
		context.info(rf'{source}stderr:')
		context.info(outputs[r'stderr'], indent=r'    ')



_warnings_regexes = (
	# doxygen
	re.compile(r'^(?P<file>.+?):(?P<line>[0-9]+): warning:\s*(?P<text>.+?)\s*$', re.I),
	# m.css
	re.compile(r'^WARNING:root:(?P<file>.+[.]xml):\s*(?P<text>.+?)\s*$', re.I),
	re.compile(r'^WARNING:root:\s*(?P<text>.+?)\s*$', re.I),
	# catch-all
	re.compile(r'^(?:Warning|Error):\s*(?P<text>.+?)\s*$', re.I)
)
_warnings_trim_suffixes = (r'Skipping it...', )
_warnings_substitutions = ((r'does not exist or is not a file', r'did not exist or was not a file'), )
_warnings_ignored = (r'inline code has multiple lines, fallback to a code block', r'libgs not found')



def extract_warnings(outputs):
	if not outputs:
		return []

	global _warnings_regexes
	global _warnings_ignored
	global _warnings_trim_suffixes
	global _warnings_substitutions

	warnings = []
	for k, v in outputs.items():
		if not v:
			continue
		output = v.split('\n')
		for o in output:
			for regex in _warnings_regexes:
				m = regex.fullmatch(o)
				if m:
					text = m[r'text'].strip()
					for suffix in _warnings_trim_suffixes:
						if text.endswith(suffix):
							text = text[:-len(suffix)].strip()
							break
					for old, new in _warnings_substitutions:
						text = text.replace(old, new)
					if not text or text in _warnings_ignored:
						break
					groups = m.groupdict()
					if r'file' in groups:
						if r'line' in groups:
							warnings.append(rf"{m[r'file']}:{m[r'line']}: {text}")
						else:
							warnings.append(rf"{m[r'file']}: {text}")
					else:
						warnings.append(text)
					break
	return warnings



def run_doxygen(context: Context):
	assert context is not None
	assert isinstance(context, Context)
	with make_temp_file() as stdout, make_temp_file() as stderr:
		try:
			subprocess.run([str(doxygen.path()), str(context.doxyfile_path)],
				check=True,
				stdout=stdout,
				stderr=stderr,
				cwd=context.input_dir)
		except:
			context.info(r'Doxygen failed!')
			dump_output_streams(context, read_output_streams(stdout, stderr), source=r'Doxygen')
			raise
		if context.is_verbose() or context.warnings.enabled:
			outputs = read_output_streams(stdout, stderr)
			if context.is_verbose():
				dump_output_streams(context, outputs, source=r'Doxygen')
			if context.warnings.enabled:
				warnings = extract_warnings(outputs)
				for w in warnings:
					context.warning(w)

	# remove the local paths from the tagfile since they're meaningless (and a privacy breach)
	if context.generate_tagfile and context.tagfile_path:
		text = read_all_text_from_file(context.tagfile_path, logger=context.verbose_logger)
		text = re.sub(r'\n\s*?<path>.+?</path>\s*?\n', '\n', text, re.S)
		context.verbose(rf'Writing {context.tagfile_path}')
		with open(context.tagfile_path, 'w', encoding='utf-8', newline='\n') as f:
			f.write(text)



def run_mcss(context: Context):
	assert context is not None
	assert isinstance(context, Context)
	with make_temp_file() as stdout, make_temp_file() as stderr:
		doxy_args = [str(context.mcss_conf_path), r'--no-doxygen', r'--sort-globbed-files']
		if context.is_verbose():
			doxy_args.append(r'--debug')
		try:
			run_python_script(
				Path(dirs.MCSS, r'documentation/doxygen.py'),
				*doxy_args,
				stdout=stdout,
				stderr=stderr,
				cwd=context.input_dir
			)
		except:
			context.info(r'm.css failed!')
			dump_output_streams(context, read_output_streams(stdout, stderr), source=r'm.css')
			raise
		if context.is_verbose() or context.warnings.enabled:
			outputs = read_output_streams(stdout, stderr)
			if context.is_verbose():
				dump_output_streams(context, outputs, source=r'm.css')
			if context.warnings.enabled:
				warnings = extract_warnings(outputs)
				for w in warnings:
					context.warning(w)



def run(
	config_path: Path = None,
	output_dir: Path = '.',
	output_html: bool = True,
	output_xml: bool = False,
	threads: int = -1,
	cleanup: bool = True,
	verbose: bool = False,
	logger=None,
	html_include: str = None,
	html_exclude: str = None,
	treat_warnings_as_errors: bool = None,
	theme: str = None,
	copy_assets: bool = True,
	temp_dir: Path = None,
	**kwargs
):

	timer = lambda desc: ScopeTimer(desc, print_start=True, print_end=context.verbose_logger)

	with Context(
		config_path=config_path,
		output_dir=output_dir,
		output_html=output_html,
		output_xml=output_xml,
		threads=threads,
		cleanup=cleanup,
		verbose=verbose,
		logger=logger,
		html_include=html_include,
		html_exclude=html_exclude,
		treat_warnings_as_errors=treat_warnings_as_errors,
		theme=theme,
		copy_assets=copy_assets,
		temp_dir=temp_dir,
		**kwargs
	) as context:

		preprocess_doxyfile(context)
		preprocess_tagfiles(context)
		preprocess_changelog(context)

		if not context.output_html and not context.output_xml:
			return

		# generate + postprocess XML in temp_xml_dir
		# (we always do this even when output_xml is false because it is required by the html)
		with timer(rf'Generating XML files with Doxygen {doxygen.version()}') as t:
			run_doxygen(context)
		with timer(r'Post-processing XML files') as t:
			if context.xml_v2:
				postprocess_xml_v2(context)
			else:
				postprocess_xml(context)
			parse_xml(context)
			clean_xml(context)

		with timer(r'Compiling regexes') as t:
			compile_regexes(context)

		# XML (the user-requested copy)
		if context.output_xml:

			with ScopeTimer(r'Copying XML', print_start=True, print_end=context.verbose_logger) as t:
				copy_tree(str(context.temp_xml_dir), str(context.xml_dir))

			# copy tagfile
			if context.generate_tagfile and context.tagfile_path:
				copy_file(context.tagfile_path, context.xml_dir, logger=context.verbose_logger)

		# HTML
		if context.output_html:

			# generate HTML with mcss
			preprocess_mcss_config(context)
			with timer(r'Generating HTML files with m.css') as t:
				run_mcss(context)

			# copy extra_files
			with ScopeTimer(r'Copying extra_files', print_start=True, print_end=context.verbose_logger) as t:
				for dest_name, source_path in context.extra_files.items():
					dest_path = Path(context.html_dir, dest_name).resolve()
					dest_path.parent.mkdir(exist_ok=True)
					copy_file(source_path, dest_path, logger=context.verbose_logger)

			# copy fonts
			if context.copy_assets:
				with ScopeTimer(r'Copying fonts', print_start=True, print_end=context.verbose_logger) as t:
					copy_tree(str(dirs.FONTS), str(Path(context.assets_dir, r'fonts')))

			# copy tagfile
			if context.generate_tagfile and context.tagfile_path:
				copy_file(context.tagfile_path, context.html_dir, logger=context.verbose_logger)

			# post-process html files
			with timer(r'Post-processing HTML files') as t:
				postprocess_html(context)



#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT
"""
Functions and types for working with C++ project graphs.
"""

import enum as _enum
from platform import node
from .utils import *

#=======================================================================================================================
# Node types
#=======================================================================================================================



class Namespace(object):
	'''Namespaces.'''
	pass



class Class(object):
	'''Classes.'''
	pass



class Struct(object):
	'''Structs.'''
	pass



class Union(object):
	'''Unions.'''
	pass



class Concept(object):
	'''C++20 `concept`.'''
	pass



class Function(object):
	'''Functions.'''
	pass



class Type(object):
	'''The type of a variable/enum/function return.'''
	pass



class Variable(object):
	'''Variables.'''
	pass



class Enum(object):
	'''Enums.'''
	pass



class EnumValue(object):
	'''Enum values.'''
	pass



class Typedef(object):
	'''Typedefs/aliases.'''
	pass



class Define(object):
	'''Preprocessor `#defines`'''
	pass



class Group(object):
	'''Groups (at the global level).'''
	pass



class MemberGroup(object):
	'''Member groups (at the class/union/struct level).'''
	pass



class Directory(object):
	'''A directory in the filesystem.'''
	pass



class File(object):
	'''A file.'''
	pass



class Page(object):
	'''A documentation page (e.g. Doxygen's `@page`).'''
	pass



class BriefDescription(object):
	'''A brief description of an element.'''
	pass



class DetailedDescription(object):
	'''A more detailed description of an element.'''
	pass



class Initializer(object):
	'''An initializer block (usually a code snippet).'''
	pass



class Paragraph(object):
	'''A paragraph.'''
	pass



class Text(object):
	'''Plain text.'''
	pass



class Reference(object):
	'''A reference to another node.'''
	pass



class ExternalResource(object):
	'''A reference to some resource outside the project (e.g. something in a tagfile).'''
	pass



class ExpositionMarkup(object):
	'''A 'leftover' node for representing miscellaneous markup in expository contexts.'''
	pass



class Friend(object):
	'''A friend relationship.'''
	pass



NODE_TYPES = {
	Namespace, Class, Struct, Union, Concept, Function, Variable, Enum, EnumValue, Typedef, Define, Group, MemberGroup,
	Directory, File, BriefDescription, DetailedDescription, Page, Initializer, Paragraph, Text, Reference,
	ExternalResource, Type, ExpositionMarkup, Friend
}
DESCRIPTION_NODE_TYPES = {BriefDescription, DetailedDescription}
EXPOSITION_NODE_TYPES = {
	*DESCRIPTION_NODE_TYPES, Page, Initializer, Paragraph, Text, Reference, ExternalResource, Type, ExpositionMarkup
}
CPP_TYPES = {Namespace, Class, Struct, Union, Concept, Function, Variable, Enum, EnumValue, Typedef, Define}
Namespace.CAN_CONTAIN = {
	Function, Class, Struct, Union, Variable, Typedef, Namespace, Concept, Enum, *DESCRIPTION_NODE_TYPES
}
Class.CAN_CONTAIN = {
	Class, Struct, Union, Function, Variable, Typedef, Enum, MemberGroup, Friend, *DESCRIPTION_NODE_TYPES
}
Struct.CAN_CONTAIN = Class.CAN_CONTAIN
Union.CAN_CONTAIN = Class.CAN_CONTAIN
Concept.CAN_CONTAIN = {Initializer, *DESCRIPTION_NODE_TYPES}
Function.CAN_CONTAIN = {Type, *DESCRIPTION_NODE_TYPES}
Variable.CAN_CONTAIN = {Type, Initializer, *DESCRIPTION_NODE_TYPES}
Enum.CAN_CONTAIN = {Type, EnumValue, *DESCRIPTION_NODE_TYPES}
EnumValue.CAN_CONTAIN = {Initializer, *DESCRIPTION_NODE_TYPES}
Typedef.CAN_CONTAIN = {Type, *DESCRIPTION_NODE_TYPES}
Define.CAN_CONTAIN = {Initializer, *DESCRIPTION_NODE_TYPES}
Group.CAN_CONTAIN = {t for t in NODE_TYPES if t not in (Page, )}
MemberGroup.CAN_CONTAIN = {t for t in Class.CAN_CONTAIN if t not in (MemberGroup, )}
Directory.CAN_CONTAIN = {Directory, File, *DESCRIPTION_NODE_TYPES}
File.CAN_CONTAIN = {
	Namespace, Class, Struct, Union, Concept, Function, Variable, Enum, Typedef, Define, Page, *DESCRIPTION_NODE_TYPES
}
Page.CAN_CONTAIN = {Paragraph, Text, Reference, *DESCRIPTION_NODE_TYPES, ExpositionMarkup}
BriefDescription.CAN_CONTAIN = {Paragraph, Text, Reference, ExpositionMarkup}
DetailedDescription.CAN_CONTAIN = {Paragraph, Text, Reference, ExpositionMarkup}
Initializer.CAN_CONTAIN = {Text, Reference, ExpositionMarkup}
Paragraph.CAN_CONTAIN = {Text, Reference, ExpositionMarkup}
Text.CAN_CONTAIN = set()
ExpositionMarkup.CAN_CONTAIN = {Paragraph, Text, Reference, ExpositionMarkup}
Type.CAN_CONTAIN = {Text, Reference}
Reference.CAN_CONTAIN = {*CPP_TYPES, Page, Group, MemberGroup, Directory, File, ExternalResource}
ExternalResource.CAN_CONTAIN = set()
Friend.CAN_CONTAIN = {Function, Class, Struct, Union, *DESCRIPTION_NODE_TYPES}



@_enum.unique
class AccessLevel(_enum.Enum):
	PRIVATE = 0
	PROTECTED = 1
	PUBLIC = 2



#=======================================================================================================================
# Nodes
#=======================================================================================================================



def _make_node_iterator(nodes, *types):
	assert types is not None

	def permissive_generator():
		nonlocal nodes
		for node in nodes:
			yield node

	if not types:
		return permissive_generator()

	for t in types:
		assert t is None or isinstance(t, bool) or t in NODE_TYPES

	def selective_generator():
		nonlocal nodes
		nonlocal types
		yield_with_no_type = False in types or None in types
		yield_with_any_type = True in types
		for node in nodes:
			if ((node.type is None and yield_with_no_type)
				or (node.type is not None and (yield_with_any_type or node.type in types))):
				yield node

	return selective_generator()



class _NullNodeIterator(object):

	def __iter__(self):
		return self

	def __next__(self):
		raise StopIteration



class GraphError(Error):
	"""Raised when a C++ graph error occurs."""
	pass



class GraphNodePropertyChanged(GraphError):
	"""Raised when an attempt is made to change an already-set property in a graph node."""
	pass



class Node(object):
	"""A single node in a C++ project graph."""

	class _Props(object):
		pass

	def __make_hierarchy_containers(self):
		if hasattr(self, r'_Node__children'):
			return
		self.__parents = []
		self.__parents_by_id = dict()
		self.__children = []
		self.__children_by_id = dict()

	def __init__(self, id: str):
		assert id is not None
		self.__id = id

	#==============
	# getters
	#==============

	def __property_get(self, name: str, out_type=None, default=None):
		assert name is not None
		value = None
		if hasattr(self, r'_Node__props'):
			value = getattr(self.__props, str(name), None)
		if value is None:
			value = default
		if value is not None and out_type is not None and not isinstance(value, out_type):
			value = out_type(value)
		return value

	@property
	def id(self) -> str:
		return self.__id

	@property
	def type(self):
		return self.__property_get(r'type')

	@property
	def type_name(self) -> str:
		t = self.type
		if t is None:
			return ''
		return t.__name__

	@property
	def qualified_name(self) -> str:
		return self.__property_get(r'qualified_name', str, r'')

	@property
	def local_name(self) -> str:
		return self.__property_get(r'local_name', str, r'')

	@property
	def definition(self) -> str:
		return self.__property_get(r'definition', str, r'')

	@property
	def static(self) -> bool:
		return self.__property_get(r'static', bool, False)

	@property
	def const(self) -> bool:
		return self.__property_get(r'const', bool, False)

	@property
	def constexpr(self) -> bool:
		return self.__property_get(r'constexpr', bool, False)

	@property
	def constinit(self) -> bool:
		return self.__property_get(r'constinit', bool, False)

	@property
	def consteval(self) -> bool:
		return self.__property_get(r'consteval', bool, False)

	@property
	def inline(self) -> bool:
		return self.__property_get(r'inline', bool, False)

	@property
	def final(self) -> bool:
		return self.__property_get(r'final', bool, False)

	@property
	def explicit(self) -> bool:
		return self.__property_get(r'explicit', bool, False)

	@property
	def noexcept(self) -> bool:
		return self.__property_get(r'noexcept', bool, False)

	@property
	def virtual(self) -> bool:
		return self.__property_get(r'virtual', bool, False)

	@property
	def mutable(self) -> bool:
		return self.__property_get(r'mutable', bool, False)

	@property
	def strong(self) -> bool:
		return self.__property_get(r'strong', bool, False)

	@property
	def access_level(self) -> AccessLevel:
		return self.__property_get(
			r'access_level', AccessLevel, AccessLevel.PRIVATE if self.type is Class else AccessLevel.PUBLIC
		)

	@property
	def text(self) -> str:
		return self.__property_get(r'text', str, r'')

	@property
	def is_paragraph(self) -> bool:
		return self.__property_get(r'is_paragraph', bool, False)

	@property
	def file(self) -> str:
		return self.__property_get(r'file', str, r'')

	@property
	def line(self) -> int:
		return self.__property_get(r'line', int, 0)

	@property
	def column(self) -> int:
		return self.__property_get(r'column', int, 0)

	@property
	def kind(self) -> str:
		return self.__property_get(r'kind', str, r'')

	@property
	def tag(self) -> str:
		return self.__property_get(r'tag', str, r'')

	@property
	def extra_attributes(self) -> typing.Sequence[typing.Tuple[str, str]]:
		return self.__property_get(r'extra_attributes', None, tuple())

	def __bool__(self) -> bool:
		return self.type is not None and bool(self.id)

	#==============
	# setters
	#==============

	def __property_set(self, name: str, out_type, value, strip_strings=False):
		assert name is not None
		# known types that have a sensible __bool__ operator can convert to None if false
		if isinstance(value, (str, Path, list, tuple, dict)):
			value = value if value else None
		# converting from strings sometimes lets us do some light parsing, as a treat
		if isinstance(value, str):
			if out_type is bool:
				if value.lower() in (r'no', r'false', r'disabled'):
					value = False
				elif value.lower() in (r'yes', r'true', r'enabled'):
					value = True
				else:
					raise GraphError(rf"Node '{self.id}' property '{name}' could not parse a boolean from '{value}'")
			elif out_type is AccessLevel:
				if value.lower() in (r'pub', r'public'):
					value = AccessLevel.PUBLIC
				elif value.lower() in (r'prot', r'protected'):
					value = AccessLevel.PROTECTED
				elif value.lower() in (r'priv', r'private'):
					value = AccessLevel.PRIVATE
				else:
					raise GraphError(rf"Node '{self.id}' property '{name}' could not parse access level from '{value}'")
				assert isinstance(value, AccessLevel)
		# None == keep whatever the current value is (no-op)
		# (None is never a valid value for a real graph attribute)
		if value is None:
			return
		if out_type is not None and not isinstance(value, out_type):
			value = out_type(value)
		if strip_strings and isinstance(value, str):
			value = value.strip()
		current = None
		has_props = hasattr(self, r'_Node__props')
		if has_props:
			current = getattr(self.__props, str(name), None)
		# it's OK if there's already a value as long as it's identical to the new one,
		# otherwise we throw so that we can detect when the source data is bad or the adapter is faulty
		# (since if a property _can_ be defined in multiple places it should be identical in all of them)
		if current is not None:
			if type(current) != type(value):
				raise GraphNodePropertyChanged(
					rf"Node '{self.id}' property '{name}' first seen with type {type(current)}, now seen with type {type(value)}"
				)
			if current != value:
				raise GraphNodePropertyChanged(
					rf"Node '{self.id}' property '{name}' first seen with value '{current}', now seen with value '{value}'"
				)
			return
		if not has_props:
			self.__props = Node._Props()
		setattr(self.__props, str(name), value)

	@type.setter
	def type(self, value):
		if value is None:
			return
		if value not in NODE_TYPES:
			raise GraphError(rf"Unknown C++ node type '{value}'")
		had_type = self.type is not None
		self.__property_set(r'type', None, value)
		if had_type != (self.type is not None):
			self.__deduce_local_name()
			if hasattr(self, r'_Node__children'):
				for child in self.__children:
					Node._check_connection(self, child)

	def __deduce_local_name(self):
		if not self.qualified_name or self.local_name or self.type is None:
			return
		if self.type in (Namespace, Class, Struct, Union, Concept, Function, Variable, Enum, EnumValue, Typedef):
			if self.qualified_name.find(r'<') != -1:  # templates might have template args with '::' so ignore them
				return
			self.local_name = tail(self.qualified_name, r'::')
		elif self.type in (Directory, File):
			self.local_name = tail(self.qualified_name, r'/')
		elif self.type is Define:
			self.local_name = self.qualified_name

	@qualified_name.setter
	def qualified_name(self, value: str):
		if value is not None and self.type in (Directory, File):
			value = str(value).strip().replace('\\', r'/').rstrip(r'/')
		self.__property_set(r'qualified_name', str, value, strip_strings=True)
		self.__deduce_local_name()

	@local_name.setter
	def local_name(self, value: str):
		self.__property_set(r'local_name', str, value, strip_strings=True)

	@definition.setter
	def definition(self, value: str):
		self.__property_set(r'definition', str, value)

	@static.setter
	def static(self, value: bool):
		self.__property_set(r'static', bool, value)

	@const.setter
	def const(self, value: bool):
		self.__property_set(r'const', bool, value)

	@constexpr.setter
	def constexpr(self, value: bool):
		self.__property_set(r'constexpr', bool, value)

	@constinit.setter
	def constinit(self, value: bool):
		self.__property_set(r'constinit', bool, value)

	@consteval.setter
	def consteval(self, value: bool):
		self.__property_set(r'consteval', bool, value)

	@inline.setter
	def inline(self, value: bool):
		self.__property_set(r'inline', bool, value)

	@final.setter
	def final(self, value: bool):
		self.__property_set(r'final', bool, value)

	@explicit.setter
	def explicit(self, value: bool):
		self.__property_set(r'explicit', bool, value)

	@noexcept.setter
	def noexcept(self, value: bool):
		self.__property_set(r'noexcept', bool, value)

	@virtual.setter
	def virtual(self, value: bool):
		self.__property_set(r'virtual', bool, value)

	@mutable.setter
	def mutable(self, value: bool):
		self.__property_set(r'mutable', bool, value)

	@strong.setter
	def strong(self, value: bool):
		self.__property_set(r'strong', bool, value)

	@access_level.setter
	def access_level(self, value: AccessLevel):
		self.__property_set(r'access_level', AccessLevel, value)

	@text.setter
	def text(self, value: str):
		self.__property_set(r'text', str, value)

	@is_paragraph.setter
	def is_paragraph(self, value: bool):
		self.__property_set(r'is_paragraph', bool, value)

	@file.setter
	def file(self, value: str):
		if value is not None:
			value = str(value).strip().replace('\\', r'/').rstrip(r'/')
		self.__property_set(r'file', str, value)

	@line.setter
	def line(self, value: int):
		self.__property_set(r'line', int, value)

	@column.setter
	def column(self, value: int):
		self.__property_set(r'column', int, value)

	@kind.setter
	def kind(self, value: str):
		self.__property_set(r'kind', str, value, strip_strings=True)

	@tag.setter
	def tag(self, value: str):
		self.__property_set(r'tag', str, value, strip_strings=True)

	@extra_attributes.setter
	def extra_attributes(self, value: typing.Sequence[typing.Tuple[str, str]]):
		self.__property_set(r'extra_attributes', None, value)

	#==============
	# children
	#==============

	@property
	def is_child(self) -> bool:
		return hasattr(self, r'_Node__children') and bool(self.__parents)

	@property
	def is_parent(self) -> bool:
		return hasattr(self, r'_Node__children') and bool(self.__children)

	def __contains__(self, node_or_id) -> bool:
		assert node_or_id is not None
		assert isinstance(node_or_id, (str, Node)) or node_or_id in NODE_TYPES
		if not hasattr(self, r'_Node__children'):
			return False
		if isinstance(node_or_id, Node):
			node_or_id = node_or_id.id
		if isinstance(node_or_id, str):
			return node_or_id in self.__children_by_id
		else:
			for c in self.__children:
				if c.type is node_or_id:
					return True
			return False

	def has_parent(self, *node_or_id_or_types) -> bool:
		assert node_or_id_or_types is not None
		if not hasattr(self, r'_Node__parents'):
			return False
		for target in node_or_id_or_types:
			if isinstance(target, Node):
				target = target.id
			if isinstance(target, str):
				return target in self.__parents_by_id
			else:
				assert target in NODE_TYPES
				for c in self.__parents:
					if c.type is target:
						return True
		return False

	def __getitem__(self, id_or_index: typing.Union[str, int]):
		assert id_or_index is not None
		assert isinstance(id_or_index, (str, int)) or id_or_index in NODE_TYPES
		if not hasattr(self, r'_Node__children'):
			return None
		if isinstance(id_or_index, str):
			try:
				return self.__children_by_id[id_or_index]
			except:
				return None
		elif isinstance(id_or_index, int):
			return self.__children[id_or_index]
		else:
			for c in self(id_or_index):
				return c
			raise KeyError(id_or_index.__name__)

	@classmethod
	def _check_connection(cls, source, dest):
		assert source is not None
		assert isinstance(source, Node)
		assert dest is not None
		assert isinstance(dest, Node)

		# self-connection is always illegal, regardless of type information
		if id(source) == id(dest):
			raise GraphError(rf"Node '{source.id}' may not connect to itself")

		# otherwise if we don't have type information the connection is 'OK'
		# (really this just means we defer the check until later)
		if source.type is None or dest.type is None:
			return

		# check basic connection rules
		if dest.type not in source.type.CAN_CONTAIN:
			raise GraphError(
				rf"{source.type_name} node '{source.id}' is not allowed to connect to {dest.type_name} nodes"
			)

		# check situations where a node must only belong to one parent of a particular set of types
		def check_single_parent(dest_types, source_types):
			nonlocal source
			nonlocal dest
			source_types = coerce_collection(source_types)
			dest_types = coerce_collection(dest_types)
			if source.type not in source_types or dest.type not in dest_types:
				return
			sum = 0
			for parent in dest(*source_types, parents=True):
				sum += 1
			if dest not in source:
				sum += 1
			if sum > 1:
				raise GraphError(
					rf"{dest.type_name} node '{dest.id}' is not allowed to be a member of more than one "
					+ rf"{{ {', '.join([t.__name__ for t in source_types])} }}"
				)

		check_single_parent(EnumValue, Enum)
		check_single_parent(Type, (Variable, Function, Enum, Typedef))
		check_single_parent(Friend, (Class, Struct, Union))
		check_single_parent(Reference, NODE_TYPES)

		# same again but in the other direction
		def check_single_child(source_types, dest_types):
			nonlocal source
			nonlocal dest
			source_types = coerce_collection(source_types)
			dest_types = coerce_collection(dest_types)
			if source.type not in source_types or dest.type not in dest_types:
				return
			sum = 0
			for child in source(*dest_types):
				sum += 1
			if dest not in source:
				sum += 1
			if sum > 1:
				raise GraphError(
					rf"{source.type_name} node '{source.id}' is not allowed to be connected to more than one "
					+ rf"{{ {', '.join([t.__name__ for t in dest_types])} }}"
				)

		check_single_child(Friend, (Class, Struct, Union, Function))
		check_single_child(Reference, NODE_TYPES)

	def add(self, child):
		assert child is not None
		assert isinstance(child, Node)

		# connecting to the same node twice is fine (no-op)
		if child in self:
			existing_child = self.__children_by_id[child.id]
			# check that identity is unique
			if id(child) != id(existing_child):
				raise GraphError(rf"Two different nodes seen with the same ID ('{child.id}')")
			return

		Node._check_connection(self, child)

		self.__make_hierarchy_containers()
		self.__children.append(child)
		self.__children_by_id[child.id] = child

		child.__make_hierarchy_containers()
		child.__parents.append(self)
		child.__parents_by_id[self.id] = self

	def __iter__(self):
		if not hasattr(self, r'_Node__children'):
			return _NullNodeIterator()
		return _make_node_iterator(self.__children)

	def __call__(self, *types, parents=False):
		if not hasattr(self, r'_Node__children'):
			return _NullNodeIterator()
		return _make_node_iterator(self.__parents if parents else self.__children, *types)

	def remove(self, child):
		assert child is not None
		assert isinstance(child, Node)

		if not hasattr(self, r'_Node__children') or child not in self or child is self:
			return

		self.__children.remove(child)
		del self.__children_by_id[child.id]

		child.__parents.remove(self)
		del child.__parents_by_id[self.id]

	def clear(self):
		if not hasattr(self, r'_Node__children'):
			return

		for child in self.__children:
			child.__parents.remove(self)
			del child.__parents_by_id[self.id]

		self.__children.clear()
		self.__children_by_id.clear()

	#==============
	# relationship queries
	#==============

	@property
	def is_class_member(self) -> bool:
		return self.has_parent(Class, Struct, Union)

	@property
	def is_class_member_variable(self) -> bool:
		return self.type is Variable and self.is_class_member

	@property
	def is_class_member_variable(self) -> bool:
		return self.type is Function and self.is_class_member

	@property
	def is_free_function(self) -> bool:
		return self.type is Function and not self.is_class_member

	@property
	def is_static_function(self) -> bool:
		return self.type is Function and self.static

	@property
	def is_friend(self) -> bool:
		return self.has_parent(Friend)

	@property
	def is_friend_function(self) -> bool:
		return self.type is Function and self.is_friend

	@property
	def is_friend_class(self) -> bool:
		return self.type in (Class, Struct, Union) and self.is_friend

	@property
	def has_friends(self) -> bool:
		return Friend in self

	#==============
	# misc
	#==============

	def copy(self, id=None, transform=None):
		node = Node(self.id if id is None else id)
		if transform is not None:
			transform(self, node)
		if hasattr(self, r'_Node__props'):
			node.__props = Node._Props()
			for key, val in self.__props.__dict__.items():
				if not hasattr(node.__props, key):
					setattr(node.__props, key, val)
		return node



#=======================================================================================================================
# Graph
#=======================================================================================================================



class Graph(object):
	"""A C++ project graph."""

	def __init__(self):
		self.__nodes: typing.Dict[str, Node]
		self.__nodes = dict()
		self.__next_unique_id = 0

	def __get_unique_id(self) -> str:
		id = rf'__graph_unique_id_{self.__next_unique_id}'
		self.__next_unique_id += 1
		return id

	def get_or_create_node(self, id: str = None, type=None, parent=None) -> Node:
		if id is None:
			id = self.__get_unique_id()
		assert id
		node = None
		if id not in self.__nodes:
			node = Node(id)
			self.__nodes[id] = node
		else:
			node = self.__nodes[id]
		node.type = type
		if parent is not None:
			parent.add(node)
		return node

	def __iter__(self):
		return _make_node_iterator(self.__nodes.values())

	def __call__(self, *types):
		return _make_node_iterator(self.__nodes.values(), *types)

	def __contains__(self, node_or_id) -> bool:
		assert node_or_id is not None
		assert isinstance(node_or_id, (str, Node)) or node_or_id in NODE_TYPES
		if isinstance(node_or_id, Node):
			node_or_id = node_or_id.id
		if isinstance(node_or_id, str):
			return node_or_id in self.__nodes
		else:
			for _, n in self.__nodes:
				if n.type is node_or_id:
					return True
			return False

	def __getitem__(self, id: str) -> Node:
		assert id is not None
		assert isinstance(id, str)
		try:
			return self.__nodes[id]
		except:
			return None

	def remove(self, *nodes: typing.Sequence[Node], filter=None):
		if filter is not None and not nodes:
			nodes = self.__nodes.values()
		prune = []
		for node in nodes:
			if node is None or node not in self:
				continue
			if filter is not None and not filter(node):
				continue
			for _, other in self.__nodes.items():
				if node is not other:
					other.remove(node)
			node.clear()
			prune.append(node)
		for node in prune:
			del self.__nodes[node.id]

	def validate(self):
		for node in self:
			if node.type is None:
				raise GraphError(rf"Node '{node.id}' is untyped")
			if node.type not in EXPOSITION_NODE_TYPES:
				if not node.qualified_name:
					raise GraphError(rf"{node.type_name} node '{node.id}' missing attribute 'qualified_name'")
				if not node.local_name:
					raise GraphError(rf"{node.type_name} node '{node.id}' missing attribute 'local_name'")

			if node.file.find('\\') != -1:
				raise GraphError(rf"{node.type_name} node '{node.id}' attribute 'file' contains back-slashes")
			if node.file.endswith(r'/'):
				raise GraphError(rf"{node.type_name} node '{node.id}' attribute 'file' ends with a forward-slash")
			if node.line < 0:
				raise GraphError(rf"{node.type_name} node '{node.id}' attribute 'line' is negative")
			if node.column < 0:
				raise GraphError(rf"{node.type_name} node '{node.id}' attribute 'column' is negative")

			if node.type in (Directory, File):
				if node.qualified_name.find('\\') != -1:
					raise GraphError(
						rf"{node.type_name} node '{node.id}' attribute 'qualified_name' contains back-slashes"
					)
				if node.qualified_name.endswith(r'/'):
					raise GraphError(
						rf"{node.type_name} node '{node.id}' attribute 'qualified_name' ends with a forward-slash"
					)
			if node.type in CPP_TYPES:
				if node.qualified_name.startswith(r'::'):
					raise GraphError(rf"{node.type_name} node '{node.id}' attribute 'qualified_name' starts with ::")
				if node.qualified_name.endswith(r'::'):
					raise GraphError(rf"{node.type_name} node '{node.id}' attribute 'qualified_name' ends with ::")
				if node.type is not EnumValue and not node.file:
					raise GraphError(rf"{node.type_name} node '{node.id}' missing attribute 'file'")

			if node.type in (EnumValue, Type, Friend):
				if not node.is_child:
					raise GraphError(rf"{node.type_name} node '{node.id}' is an orphan")

			if node.type in (Function, Variable, Typedef):
				if Type not in node:
					raise GraphError(rf"{node.type_name} node '{node.id}' is missing a Type")

	def copy(self, filter=None, id_transform=None, transform=None):
		g = Graph()
		id_remap = dict()
		# first pass to copy
		for src in self:
			if filter is not None and not filter(src):
				continue
			id = src.id
			if id_transform is not None:
				id = id_transform(src)
			if id is None:
				id = g.__get_unique_id()
			else:
				id = str(id)
			if id in g:
				raise GraphError(rf"A node with id '{id}' already exists in the destination graph")
			id_remap[src.id] = id
			g.__nodes[id] = src.copy(id=id, transform=transform)
		# second pass to link hierarchy
		for src in self:
			if src.id not in id_remap:
				continue
			for child in src:
				if child.id not in id_remap:
					continue
				g[id_remap[src.id]].add(g[id_remap[child.id]])
		return g
